#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 29 18:02:41 2024

@author: ncro8394
"""
from itertools import chain
from os import listdir, mkdir, path
from numpy import char, reshape
from pandas import DataFrame, read_csv, read_excel
from wonambi import Dataset
from wonambi.attr import Annotations, create_empty_annotations
from .logs import create_logger

def read_tracking_sheet(filepath, logger):
        
        track_file = [x for x in listdir(filepath) if 'tracking' in x]
        
        if len(track_file) > 1:
            logger.error('>1 tracking file found.')
            logger.warning("Is the tracking file currently open?")
            return 'error'
        elif len(track_file) == 0:
            logger.error('No tracking file found.')
            return 'error'
        else:
            track_file = track_file[0]
            if '.tsv' in track_file:
                track = read_csv(f'{filepath}/{track_file}' , sep='\t')
            elif '.xls' in track_file:
                track = read_excel(f'{filepath}/{track_file}')
    
        return track

def select_input_dirs(outpath, xml_dir, evt_name=None, 
                      logger = create_logger('Select inputs')):
    if not xml_dir:
        if evt_name in ['spindle', 'Ferrarelli2007', 'Nir2011', 'Martin2013', 
                        'Moelle2011', 'Wamsley2012', 'Ray2015', 'Lacourse2018', 
                        'FASST', 'FASST2', 'Concordia','UCSD','spindle_adap', 
                        'Ferrarelli2007_adap', 'Nir2011_adap', 'Martin2013_adap', 
                        'Moelle2011_adap', 'Wamsley2012_adap', 'Ray2015_adap', 
                        'Lacourse2018_adap', 'FASST_adap', 'FASST2_adap', 
                        'Concordia_adap','UCSD_adap']:
            xml_dir = f'{outpath}/spindle'
            
        elif evt_name in ['Ngo2015','Staresina2015','Massimini2004','slowwave',
                          'slowosc','SO']:
            xml_dir = f'{outpath}/slowwave'
            
        elif evt_name in ['event_pac']:
            xml_dir = f'{outpath}/event_pac'
            
        elif evt_name in ['pac']:
            xml_dir = f'{outpath}/pac' 
            
        elif evt_name in ['cluster']:
            xml_dir = f'{outpath}/clusterfluc'
            
        elif evt_name in ['staging', 'macro', None]:
            xml_dir = [x for x in listdir(outpath) if 'staging' in x]
            
            if len(xml_dir) > 1:
                xml_dir = [x for x in xml_dir if 'manual' in x]
                logger.warning(f">1 derivatives directories have 'staging' in the "
                               f"name. Will default to {xml_dir}")
                
            elif len(xml_dir) == 1:
                xml_dir = f'{outpath}/{xml_dir[0]}'

            else:
                logger.error(" 'xml_dir' wasn't specified and it cannot be "
                                "determined from inside the derivatives directory. "
                                "Please specify manually.") 
                xml_dir = ''
        else:
            xml_dir = f'{outpath}/{evt_name}'
        
    return xml_dir

def select_output_dirs(outpath, out_dir, evt_name=None, 
                      logger = create_logger('Select outputs')):
            
    if not out_dir:
        if evt_name in ['spindle', 'Ferrarelli2007', 'Nir2011', 'Martin2013', 
                        'Moelle2011', 'Wamsley2012', 'Ray2015', 'Lacourse2018', 
                        'FASST', 'FASST2', 'Concordia','UCSD','spindle_adap', 
                        'Ferrarelli2007_adap', 'Nir2011_adap', 'Martin2013_adap', 
                        'Moelle2011_adap', 'Wamsley2012_adap', 'Ray2015_adap', 
                        'Lacourse2018_adap', 'FASST_adap', 'FASST2_adap', 
                        'Concordia_adap','UCSD_adap']:
            out_dir = f'{outpath}/spindle'
        elif evt_name in ['Ngo2015','Staresina2015','Massimini2004','slowwave',
                          'slowosc','SO']:
            out_dir = f'{outpath}/slowwave'
        elif evt_name in ['pac']:
            out_dir = f'{outpath}/pac'
        elif evt_name in ['cluster']:
            out_dir = f'{outpath}/clusterfluc'
            
        elif evt_name in ['staging', 'macro', None]:
            out_dir = [x for x in listdir(outpath) if 'staging' in x]
            
            if len(out_dir) > 1:
                out_dir = [x for x in out_dir if 'manual' in x]
                logger.warning(f">1 derivatives directories have 'staging' in "
                               f"the name. Will default to {out_dir}")
            if len(out_dir) == 1:
                out_dir = f'{outpath}/{out_dir[0]}'

            else:
                logger.error(" 'xml_dir' wasn't specified and it cannot be "
                                "determined from inside the derivatives directory. "
                                "Please specify manually.") 
                xml_dir = ''    
        else:
            out_dir = f'{outpath}/{evt_name}'
    
    if not path.exists(out_dir):
        mkdir(out_dir)
        
    return out_dir

def load_stages(in_dir, xml_dir, subs = 'all', sessions = 'all', filetype = '.edf',
                stage_key = None, logger = create_logger('Load stages')):
    
    '''
        Extracts stages from the BIDS formatted dataset, in which
        staging has been listed in a file *acq-PSGScoring_events.tsv, and
        saves the information in an annotations file
        
        Parameters
        ----------
                    in_dir   :  str
                                The path to the BIDS dataset containing EEG recordings
                                and staging *acq-PSGScoring_events.tsv files
                    xml_dir  :  str
                                The derivatives path to store the annotations (.xml) 
                                file under the <sub>/<ses> structure
                    subs     :  str or ist of str
                                The participant ids to run this function on. Can 
                                be set to 'all', and all participants will be 
                                formatted.
                    sessions :  str or list of str
                                The participant ids to run this function on. Can 
                                be set to 'all', and all participants will be 
                                formatted.   
                    filetype :  str
                                The extension of EEG recording files 
                                (default = '.edf')
                    stage_key : dict or NoneType
                                Key for staging names to be saved into annotations
                                file (default is set to be compatible with Wonambi)
                    logger   :  logger for logging
                    

        Returns
        -------
                    flag : 0 (no errors)
                           1+ (errors or warnings)
        
        
    '''
    
    flag = 0
    if not stage_key:
        stage_key = {0: 'Wake',
                     1: 'NREM1',
                     2: 'NREM2',
                     3: 'NREM3',
                     4: 'REM',
                     9: 'Undefined'}
    
    # Define subs    
    if subs == 'all':
        subs = [x for x in listdir(in_dir) if '.' not in x]
    elif not isinstance(subs, list):  
        logger.error(f"'subs' must be a list of subject ids or 'all', not {subs}.")
        return 
    
    # Loop through
    subs.sort()
    for s, sub in enumerate(subs):
        
        # Make subject output directory
        if not path.exists(f'{xml_dir}/{sub}'):
            mkdir(f'{xml_dir}/{sub}')
            
        # Define sessions
        subdir = f'{in_dir}/{sub}'
        if sessions == 'all':
            sess = [x for x in listdir(subdir) if '.' not in x]
        elif not isinstance(sessions, list):  
            logger.error(f"'sessions' must be a list of session ids or 'all', not {sessions}.")
            return
        else:
            sess = sessions
        
        # flag, sess = load_sessions(sub, sessions, in_dir, flag, 
        #                          logger, verbose=2)
        
        # Loop through
        for s, ses in enumerate(sess):
            
            # Make session output directory
            if not path.exists(f'{xml_dir}/{sub}/{ses}'):
                mkdir(f'{xml_dir}/{sub}/{ses}')
                
            # Load BIDS stage event file
            datadir = f'{in_dir}/{sub}/{ses}/eeg'
            stagefile = [x for x in listdir(datadir) if 'acq-PSGScoring_events.tsv' in x
                         if not x.startswith('.')][0]
            stagedf = read_csv(f'{datadir}/{stagefile}', sep ='\t' ) 
            
            # Create new annotations for staging
            xml_file = f'{xml_dir}/{sub}/{ses}/{sub}_{ses}_staging.xml'
            if not path.exists(xml_file):
                edf_file = [x for x in listdir(datadir) if x.endswith(filetype)][0]
                dset = Dataset(f'{datadir}/{edf_file}')
                create_empty_annotations(xml_file, dset)
                logger.debug(f'Creating annotations file for {sub}, {ses}')
            else:
                logger.warning(f"Annotations file exists for {sub}, {ses}, staging "
                               "will be overwritten.")
                flag += 1
            annot = Annotations(xml_file)
            
            # Save staging to annotations
            annot.add_rater('manual')
            stage_flag = 0
            for i in range(0, stagedf.shape[0]):
                epoch_beg = stagedf.loc[i, 'onset']
                try:
                    one_stage = stage_key[stagedf.loc[i, 'staging']]
                except:
                    one_stage = 'Undefined'
                try:
                    annot.set_stage_for_epoch(epoch_beg, one_stage,
                                         attr='stage',
                                         save=False)
                except:
                    stage_flag += 1
            if stage_flag > 0:
                logger.warning("Some epochs in staging file were outside the length "
                               "of the edf. It looks as though the staging file is "
                               "not the same length as the edf for {sub}, {ses}. REVIEW.")
            annot.save()
    
    return flag

def check_chans(datapath, chan, ref_chan, logger):
    if chan is None:
        chan = read_tracking_sheet(f'{datapath}', logger)
        if not isinstance(chan, DataFrame) and chan == 'error':
            logger.error("Channels haven't been defined, and there was an error "
                         "reading the tracking file.")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            return chan, ref_chan
        
    if ref_chan is None:
        ref_chan = read_tracking_sheet(f'{datapath}', logger)
        if not isinstance(ref_chan, DataFrame) and ref_chan == 'error':
            logger.warning("Reference channels haven't been defined, and there "
                           "was an error reading the tracking file.")
            logger.info('')
            logger.info("Check documentation for how to set up channel data: "
                        "https://seapipe.readthedocs.io/en/latest/index.html")
            logger.info('-' * 10)
            logger.warning('No re-referencing will be performed prior to analysis.')
            ref_chan = None
    
    if ref_chan is False:
        return chan
    else:
        return chan, ref_chan

def load_sessions(sub, ses, rec_dir = None, flag = 0, 
                  logger = create_logger("Load Sessions"), verbose = 2):
    
    ''' Function to load in sessions available for a subject from a tracking 
        sheet.
    
        Parameters
        ----------
                sub : str
                      The subject id to load channel information for
                ses : str
                      The session id to load channel information for. Can be 
                      set as 'all' if all sessions for that subject are 
                      desired.
                rec_dir : str or None (optional)
                      The path to the BIDS dataset. If ses = 'all', this MUST
                      be specified.
                flag : int
                       the status of error logging
                logger : Class of logging.Logger
                         To log output (of course)
                         see: https://docs.python.org/3/library/logging.html
                verbose : int
                        The verbosity of logging.
                        verbose = 0 (error only)
                        verbose = 1 (warning only)
                        verbose = 2 (debug mode)
        
        Returns
        -------
                flag : int
                       whether an error occured in loading channel names
                
                chanset : dict
                         Format 'chan_name':'ref_chan' specifying the 
                         channel and it's associated reference channels.
    '''
    
    if type(ses) == type(DataFrame()):
        if verbose==2:
            logger.debug("Reading channel names from tracking file")
        # Search participant
        sub_row = ses[ses['sub']==sub]
        if sub_row.size == 0:
            if verbose>0:
                logger.warning(f"Participant {sub} not found in column 'sub' in "
                               "tracking file.")
                flag+=1
                return flag, None
        ses = [x for x in sub_row['ses']]
    elif type(ses) == str and ses == 'all':
        if not rec_dir:
            logger.error("If loading sessions from tracking sheet with ses = 'all', "
                         "rec_dir MUST be specified.")
            ses = None
            flag +=1
        else:   
            ses = listdir(rec_dir + '/' + sub)
            ses = [x for x in ses if not '.' in x]
    elif not type(ses) == list:
        logger.error("'sessions' must be set to None,'all' or a list of sub ids. "
                     "For session setup options, refer to documentation:")
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
        flag+=1
        return flag, None
    
    return flag, ses


def load_channels(sub, ses, chan, ref_chan, flag = 0, 
                  logger = create_logger("Load Channels"), verbose = 2):
    
    ''' Function to load in channels and their references from a tracking sheet.
    
        Parameters
        ----------
                sub : str
                      The subject id to load channel information for
                ses : str
                      The session id to load channel information for
                chan : list or DataFrame
                      The channel names to either load (DataFrame) or 
                      associate with a reference (list). If a DataFrame, 
                      the columns 'sub', 'ses', and 'chanset' must exist.
                ref_chan : list or DataFrame
                      The ref channel names to either load (DataFrame) or 
                      associate with a reference (list). If a DataFrame, 
                      the columns 'sub', 'ses' and 'refset' must exist.
                flag : int
                       the status of error logging
                logger : Class of logging.Logger
                         To log output (of course)
                         see: https://docs.python.org/3/library/logging.html
                verbose : int
                        The verbosity of logging.
                        verbose = 0 (error only)
                        verbose = 1 (warning only)
                        verbose = 2 (debug mode)
        
        Returns
        -------
                flag : int
                       whether an error occured in loading channel names
                
                chanset : dict
                         Format 'chan_name':'ref_chan' specifying the 
                         channel and it's associated reference channels.
    '''
    
    # verbose = 0 (error only)
    # verbose = 1 (warning only)
    # verbose = 2 (debug)
    
    # 1. Extract relevant cells from tracking sheet
    if isinstance(chan, DataFrame):
        if verbose==2:
            logger.debug("Reading channel names from tracking file")
        # Search participant
        chans = chan[chan['sub']==sub]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Participant {sub} not found in column 'sub' in"
                               f" tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search session
        chans = chans[chans['ses']==ses]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Session {ses} not found in column 'ses' in "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Extract channels
        chans = search_chans(chans)
        chans = chans.dropna(axis=1, how='all')
        if len(chans.columns) == 0:
            if verbose>0:
                logger.warning("No channel set found in tracking file for "
                               f"{sub}, {ses}, skipping...")
            flag+=1
            return flag, None
    else:
        chans = chan
    
    if isinstance(ref_chan, DataFrame):
        if verbose==2:
            logger.debug("Reading reference channel names from tracking file ")
        ref_chans = ref_chan[ref_chan['sub']==sub]
        if ref_chans.size == 0:
            if verbose>0:
                logger.warning("Participant not found in column 'sub' in "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        ref_chans = ref_chans[ref_chans['ses']==ses]
        if ref_chans.size == 0:
            if verbose>0:
                logger.warning("Session not found in column 'ses' in tracking "
                               f"file for {sub}, {ses}.")
            flag+=1
            return flag, None
        ref_chans = ref_chans.filter(regex='refset')
        ref_chans = ref_chans.dropna(axis=1, how='all')
        if len(ref_chans.columns) == 0:
            if verbose>0:
                logger.warning("No reference channel set found in tracking "
                               f"file for {sub}, {ses}. Progressing without "
                               "re-referencing...")
            ref_chans = []
    elif isinstance(ref_chan, list):
        ref_chans = ref_chan
    else:
        ref_chans = []
    
    if isinstance(chans, list):
        if isinstance(ref_chans, DataFrame) and len(ref_chans.columns) >1:
            chan = ref_chan[ref_chan['sub']==sub]
            chan = chan[chan['ses']==ses]
            chan = search_chans(chan)
            ref_chan=[]
            for c in chans:
                ref_chan.append([ref_chans[ref_chans.columns[x]].iloc[0] for 
                                 x, y in enumerate(chan) if c in chan[y].iloc[0]][0])
            ref_chan = [char.split(x, sep=', ').tolist() for x in ref_chan]
            chanset = {chn:[ref_chan[i]] if isinstance(ref_chan[i],str) else 
                       ref_chan[i] for i,chn in enumerate(chans)}
            
        elif isinstance(ref_chans, DataFrame):
            ref_chans = ref_chans.to_numpy()[0]
            ref_chans = ref_chans.astype(str)
            ref_chans_all = []
            for cell in ref_chans:
                cell = cell.split(', ')
                for x in cell:
                    if ',' in x: 
                        x = x.split(',')
                        ref_chans_all = ref_chans_all + x
                    else:
                        ref_chans_all.append(x)
            ref_chans = [x for x in ref_chans_all if not x=='']
            chanset = {chn:ref_chans for chn in chans} 
        
        elif isinstance(ref_chans, list):
            chanset = {chn:ref_chans for chn in chans}
        else:
            chanset = {chn:[] for chn in chans}
    
    elif isinstance(chans, DataFrame):
        if type(ref_chans) == DataFrame and len(ref_chans.columns) != len(chans.columns):
            logger.error("There must be the same number of channel sets and "
                         "reference channel sets in 'tracking file, but for "
                         f"{sub}, {ses}, there were {len(chans.columns)} channel "
                         f"sets and {len(ref_chans.columns)} reference channel sets. "
                         "For channel setup options, refer to documentation:")
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            flag+=1
            return flag, None
        elif isinstance(ref_chans, DataFrame):
            ref_chans = ref_chans.to_numpy()[0]
            ref_chans = ref_chans.astype(str)
            ref_chans_all = []
            for cell in ref_chans:
                cell = cell.split(', ')
                refcell = []
                for x in cell:
                    if ',' in x: 
                        x = x.split(',')
                        refcell = refcell + x
                    else:
                        refcell.append(x)
                refcell = [x for x in refcell if not x=='']
                ref_chans_all.extend(refcell)
            ref_chans = [x for x in ref_chans_all if not x=='']
        
        chans = chans.to_numpy()[0]
        chans = chans.astype(str)
        chans_all = []
        for cell in chans:
            cell = cell.split(', ')
            chancell = []
            for x in cell:
                if ',' in x: 
                    x = x.split(',')
                    chancell = chancell + x
                else:
                    chancell.append(x)
            chancell = [x for x in chancell if not x=='']
            chans_all.append(chancell)
        chans = chans_all

        if len(ref_chans) > 0:
            chanset = {key: ref_chans for chn in chans for key in chn}
        else:
            chanset = {key: [] for chn in chans for key in chn}
        
    else:
        logger.error("The variable 'chan' should be a [list] or definied in the "
                     "'chanset' column of tracking file - NOT a string.")
        flag+=1
        return flag, None
    
    return  flag, chanset


def rename_channels(sub, ses, chan, logger):
    
    if type(chan) == type(DataFrame()):
        # Search participant
        chans = chan[chan['sub']==sub]
        if len(chans.columns) == 0:
            return None
        # Search session
        chans = chans[chans['ses']==ses]
        if len(chans.columns) == 0:
            return None
        # Search channels
        oldchans = search_chans(chans)
        chans = chans.filter(regex='^((?!invert).)*$')
        oldchans = oldchans.dropna(axis=1, how='all')
        if len(oldchans.columns) == 0:
            return None
        newchans = chans.filter(regex='rename')
        newchans = newchans.dropna(axis=1, how='all')
        if len(newchans.columns) == 0:
            return None
    else:
        return None  
    
    if isinstance(oldchans,DataFrame):
        if isinstance(newchans,DataFrame) and len(newchans.columns) != len(oldchans.columns):
            try:
                oldchans_to_be_renamed = oldchans[list({i for i in oldchans if any(i in j for j in newchans)})]
                oldchans_to_be_kept = oldchans[list({i for i in oldchans if not any(i in j for j in newchans)})]
            except:
                logger.warning(f"There must be the same number of channel sets "
                               "and channel rename sets in tracking file, but "
                               f"for {sub}, {ses}, there were {len(oldchans.columns)} "
                               f"channel sets and {len(newchans.columns)} channel "
                               "rename sets. For info on how to rename channels, "
                               "refer to documentation:")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.warning(f"Using original channel names for {sub}, {ses}...")
                return None
        else:
            oldchans_to_be_renamed = oldchans
            oldchans_to_be_kept = oldchans
        
        # Split cells in tracking
        # OLD Channels to be renamed
        oldchans_to_be_renamed = oldchans_to_be_renamed.to_numpy() 
        oldchans_to_be_renamed = oldchans_to_be_renamed[0].astype(str)
        oldchans_all = []
        for cell in oldchans_to_be_renamed:
            cell = cell.split(', ')
            chancell = []
            for x in cell:
                if ',' in x: 
                    x = x.split(',')
                    chancell = chancell + x
                else:
                    chancell.append(x)
            chancell = [x for x in chancell if not x=='']
            oldchans_all.append(chancell)       
        oldchans_to_be_renamed = list(chain(*oldchans_all))
        
        # OLD Channels to be kept
        oldchans_to_be_kept = oldchans_to_be_kept.to_numpy() 
        oldchans_to_be_kept = oldchans_to_be_kept[0].astype(str)
        oldchans_all = []
        for cell in oldchans_to_be_kept:
            cell = cell.split(', ')
            chancell = []
            for x in cell:
                if ',' in x: 
                    x = x.split(',')
                    chancell = chancell + x
                else:
                    chancell.append(x)
            chancell = [x for x in chancell if not x=='']
            oldchans_all.append(chancell)       
        oldchans_to_be_kept = list(chain(*oldchans_all))
        
        if type(newchans) == DataFrame:
            newchans = newchans.to_numpy() 
            newchans = newchans[0].astype(str)
            newchans_all = []
            for cell in newchans:
                cell = cell.split(', ')
                chancell = []
                for x in cell:
                    if ',' in x: 
                        x = x.split(',')
                        chancell = chancell + x
                    else:
                        chancell.append(x)
                chancell = [x for x in chancell if not x=='']
                newchans_all.append(chancell)       
            newchans = list(chain(*newchans_all))
        
        if len(oldchans_to_be_renamed) == len(newchans):
            newchans = {chn:newchans[i] for i,chn in enumerate(oldchans_to_be_renamed)}
            n = {x:x for x in oldchans_to_be_kept}
            newchans = n | newchans
        else:
            logger.warning(f"There must be the same number of original channel "
                           f"names and new renamed channels in tracking file, "
                           f"but for {sub}, {ses}, there were {len(oldchans)} old "
                           f"channel and {len(newchans)} new channel names. For "
                           "info on how to rename channels, refer to documentation:")
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.warning(f"Using original channel names for {sub}, {ses}...")
            return None
    else:
        return None
    
    return newchans

def reverse_chan_lookup(sub, ses, chan, logger):
    
    if type(chan) == type(DataFrame()):
        # Search participant
        chans = chan[chan['sub']==sub]
        if len(chans.columns) == 0:
            return None
        # Search session
        chans = chans[chans['ses']==ses]
        if len(chans.columns) == 0:
            return None
        # Search channels
        newchans = search_chans(chans)
        chans = chans.filter(regex='^((?!invert).)*$')
        newchans = newchans.dropna(axis=1, how='all')
        if len(newchans.columns) == 0:
            return None
        oldchans = chans.filter(regex='rename')
        oldchans = oldchans.dropna(axis=1, how='all')
        if len(oldchans.columns) == 0:
            return None
    else:
        return None  
    
    if isinstance(oldchans,DataFrame):
        if isinstance(newchans,DataFrame) and len(newchans.columns) != len(oldchans.columns):
            try:
                oldchans_to_be_renamed = oldchans[list({i for i in oldchans if any(i in j for j in newchans)})]
                oldchans_to_be_kept = oldchans[list({i for i in oldchans if not any(i in j for j in newchans)})]
            except:
                logger.warning(f"There must be the same number of channel sets "
                               "and channel rename sets in tracking file, but "
                               f"for {sub}, {ses}, there were {len(oldchans.columns)} "
                               f"channel sets and {len(newchans.columns)} channel "
                               "rename sets. For info on how to rename channels, "
                               "refer to documentation:")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.warning(f"Using original channel names for {sub}, {ses}...")
                return None
        else:
            oldchans_to_be_renamed = oldchans
            oldchans_to_be_kept = oldchans
        
        # Split cells in tracking
        # OLD Channels to be renamed
        oldchans_to_be_renamed = oldchans_to_be_renamed.to_numpy() 
        oldchans_to_be_renamed = oldchans_to_be_renamed[0].astype(str)
        oldchans_all = []
        for cell in oldchans_to_be_renamed:
            cell = cell.split(', ')
            chancell = []
            for x in cell:
                if ',' in x: 
                    x = x.split(',')
                    chancell = chancell + x
                else:
                    chancell.append(x)
            chancell = [x for x in chancell if not x=='']
            oldchans_all.append(chancell)       
        oldchans_to_be_renamed = list(chain(*oldchans_all))
        
        # OLD Channels to be kept
        oldchans_to_be_kept = oldchans_to_be_kept.to_numpy() 
        oldchans_to_be_kept = oldchans_to_be_kept[0].astype(str)
        oldchans_all = []
        for cell in oldchans_to_be_kept:
            cell = cell.split(', ')
            chancell = []
            for x in cell:
                if ',' in x: 
                    x = x.split(',')
                    chancell = chancell + x
                else:
                    chancell.append(x)
            chancell = [x for x in chancell if not x=='']
            oldchans_all.append(chancell)       
        oldchans_to_be_kept = list(chain(*oldchans_all))
        
        if type(newchans) == DataFrame:
            newchans = newchans.to_numpy() 
            newchans = newchans[0].astype(str)
            newchans_all = []
            for cell in newchans:
                cell = cell.split(', ')
                chancell = []
                for x in cell:
                    if ',' in x: 
                        x = x.split(',')
                        chancell = chancell + x
                    else:
                        chancell.append(x)
                chancell = [x for x in chancell if not x=='']
                newchans_all.append(chancell)       
            newchans = list(chain(*newchans_all))
        
        if len(oldchans_to_be_renamed) == len(newchans):
            newchans = {chn:newchans[i] for i,chn in enumerate(oldchans_to_be_renamed)}
            n = {x:x for x in oldchans_to_be_kept}
            newchans = n | newchans
        else:
            logger.warning(f"There must be the same number of original channel "
                           f"names and new renamed channels in tracking file, "
                           f"but for {sub}, {ses}, there were {len(oldchans)} old "
                           f"channel and {len(newchans)} new channel names. For "
                           "info on how to rename channels, refer to documentation:")
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.warning(f"Using original channel names for {sub}, {ses}...")
            return None
    else:
        return None
    
    return newchans

def load_stagechan(sub, ses, chan, ref_chan, flag, logger, verbose=2):
    # verbose = 0 (error only)
    # verbose = 1 (warning only)
    # verbose = 2 (debug)
    
    if type(chan) == type(DataFrame()):
        if verbose==2:
            logger.debug("Reading channel names from tracking file")
        # Search participant
        chans = chan[chan['sub']==sub]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Participant {sub} not found in column 'sub' in "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search session
        chans = chans[chans['ses']==ses]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Session {ses} not found in column 'ses' in "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search channel
        chans = chans.filter(regex='stagechan')
        chans = chans.dropna(axis=1, how='all')
        if len(chans.columns) == 0:
            if verbose>0:
                logger.warning(f"No stagechan found in tracking file for {sub}, "
                               f"{ses}, skipping...")
            flag+=1
            return flag, None
    else:
        chans = chan
    
    if type(ref_chan) == type(DataFrame()):
        if verbose==2:
            logger.debug("Reading reference channel names from tracking file")
        ref_chans = ref_chan[ref_chan['sub']==sub]
        if ref_chans.size == 0:
            if verbose>0:
                logger.warning(f"Participant not found in column 'sub' in tracking "
                               f"file for {sub}, {ses}.")
            flag+=1
            return flag, None
        ref_chans = ref_chans[ref_chans['ses']==ses]
        if ref_chans.size == 0:
            if verbose>0:
                logger.warning(f"Session not found in column 'ses' in tracking "
                               f"file for {sub}, {ses}.")
            flag+=1
            return flag, None
        ref_chans = ref_chans.filter(regex='refset')
        ref_chans = ref_chans.dropna(axis=1, how='all')
        if len(ref_chans.columns) == 0:
            if verbose>0:
                logger.warning(f"No reference channel set found in tracking file "
                               f"for {sub}, {ses}. Progressing without re-referencing...")
            ref_chans = []
    elif ref_chan:
        ref_chans = ref_chan
    else:
        ref_chans = []
    
    if type(chans) == list:
        if type(ref_chans) == DataFrame and len(ref_chans.columns) >1:
            chan = ref_chan[ref_chan['sub']==sub]
            chan = chan[chan['ses']==ses]
            chan = search_chans(chan)
            ref_chan=[]
            for c in chans:
                ref_link = [ref_chans[ref_chans.columns[x]].iloc[0] for x, y 
                                           in enumerate(chan) if c in chan[y].iloc[0]]
                if len(ref_link) > 1:
                    ref_chan.append(ref_link[0])
            ref_chan = [char.split(x, sep=', ').tolist() for x in ref_chan]
            
            if len(ref_chan) > 0:
                chanset = {chn:[ref_chan[i]] if isinstance(ref_chan[i],str) 
                           else ref_chan[i] for i,chn in enumerate(chans)}
            else:
                chanset = {chn:ref_chan for chn in chans}
            
        elif type(ref_chans) == DataFrame:
            ref_chans = ref_chans.to_numpy()[0]
            ref_chans = ref_chans.astype(str)
            ref_chans = char.split(ref_chans, sep=', ')
            ref_chans = [x for y in ref_chans for x in y]
            chanset = {chn:ref_chans for chn in chans}    
        else:
            chanset = {chn:[] for chn in chans}
    
    elif type(chans) == type(DataFrame()):
        
        if type(ref_chans) == DataFrame:
            if len(ref_chans.columns) != len(chans.columns):
                logger.warning(f"There were >2 reference channel sets in 'tracking' "
                               f"file for {sub}, {ses}, we will just use the "
                               "first set for automatic staging.")
                ref_chans = ref_chans.iloc[:,0]
                
            ref_chans = ref_chans.to_numpy()[0]
            if not isinstance(ref_chans, str):
                ref_chans = ref_chans.astype(str)
            ref_chans = char.split(ref_chans, sep=', ')
            if ref_chans.size < 2:
                ref_chans = reshape(ref_chans, (1,1))
                ref_chans = [x for x in ref_chans[0][0]]
            else:
                ref_chans = [x for x in ref_chans]

        chans = chans.to_numpy()[0]
        chans = chans.astype(str)
        chans = char.split(chans, sep=', ')
        chans = [x for x in chans]
        if len(ref_chans)>0:
            chanset = {key:ref_chans[i] for i,chn in enumerate(chans) for key in chn}
        else:
            chanset = {key:[] for i,chn in enumerate(chans) for key in chn}
        
    else:
        logger.error("The variable 'chan' should be definied in the 'chanset' "
                     "column of tracking file or a [list] - NOT a string.")
        flag+=1
        return flag, None
    
    return  flag, chanset


def load_eog(sub, ses, chan, flag, logger, verbose=2):
    # verbose = 0 (error only)
    # verbose = 1 (warning only)
    # verbose = 2 (debug)
    
    if type(chan) == type(DataFrame()):
        if verbose==2:
            logger.debug("Reading eog channel names from tracking file")
        # Search participant
        chans = chan[chan['sub']==sub]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Participant not found in column 'sub' in "
                               f"the tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search session
        chans = chans[chans['ses']==ses]
        if chans.size == 0:
            if verbose>0:
                logger.warning(f"Session not found in column 'ses' in the tracking "
                               f"file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search channel
        chans = chans.filter(regex='eog')
        chans = chans.dropna(axis=1, how='all')
        if len(chans.columns) == 0:
            if verbose>0:
                logger.warning("No stagechan found in tracking file for "
                               f"{sub}, {ses}...")
            flag+=1
            return flag, []
        chans = [x for x in chans['eog']]
    else:
        chans = chan
    return flag, chans


def load_emg(sub, ses, chan, flag, logger, verbose=2):
    # verbose = 0 (error only)
    # verbose = 1 (warning only)
    # verbose = 2 (debug)
    
    if type(chan) == type(DataFrame()):
        if verbose==2:
            logger.debug("Reading emg channel names from tracking file")
        # Search participant
        chans = chan[chan['sub']==sub]
        if chans.size == 0:
            if verbose>0:
                logger.warning("Participant not found in column 'sub' in the "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search session
        chans = chans[chans['ses']==ses]
        if chans.size == 0:
            if verbose>0:
                logger.warning("Session not found in column 'ses' in the "
                               f"tracking file for {sub}, {ses}.")
            flag+=1
            return flag, None
        # Search channel
        chans = chans.filter(regex='emg')
        chans = chans.dropna(axis=1, how='all')
        if len(chans.columns) == 0:
            if verbose>0:
                logger.warning(f"No emg found in tracking file for {sub}, {ses}...")
            flag+=1
            return flag, []
        chans = [x for x in chans['emg']]
    else:
        chans = chan
    return flag, chans

def check_adap_bands(rootpath, subs, sessions, chan, logger):
    
    try:
        track = read_tracking_sheet(rootpath, logger)
    except:
        logger.error("Error reading tracking sheet. Check that it isn't open.")
        logger.info("For info how to use adap_bands = 'Manual' in detections, "
                    "refer to documentation:")
        logger.info(" https://seapipe.readthedocs.io/en/latest/index.html")
        logger.info('-' * 10)
        return 'error'
    
    chans = search_chans(track)
    chans = chans.dropna(axis=1, how='all')
    peaks = track.filter(regex='peaks')
    peaks = peaks.dropna(axis=1, how='all')
    
    if len(peaks.columns) == 0:
        logger.error("No spectral peaks have been provided in tracking file. "
                     "Peaks will need to be detected.")
        logger.info("Check documentation for how to use adap_bands = 'Manual' "
                    "in detections: https://seapipe.readthedocs.io/en/latest/index.html")
        logger.info('-' * 10)
        return 'error'
    elif len(peaks.columns) != len(chans.columns):
        logger.error("There must be the same number of channel sets and spectral "
                     "peaks sets in tracking file")
        logger.info("Check documentation for how to use adap_bands = 'Manual' "
                    "in detections: https://seapipe.readthedocs.io/en/latest/index.html")
        return 'error'
    
    sub = {}
    for c, col in enumerate(chans.columns):
        for r, row in enumerate(chans[col]):
            chs = reshape(char.split(str(row), sep=', '), (1,1))[0][0]
            pks = reshape(char.split(str(peaks.iloc[r,c]), sep=', '), (1,1))[0][0]  
            if len(chs) != len(pks) and 'nan' not in (pks):
                logger.warning(f"For {track['sub'][r]}, {track['ses'][r]} the "
                               f"number of channels provided ({len(chs)}) != the "
                               f"number of spectral peaks ({len(pks)}).")
                if not track['sub'][r] in sub.keys():
                    sub[track['sub'][r]] = [track['ses'][r]]
                else:
                    sub[track['sub'][r]].append(track['ses'][r])
            elif 'nan' in (pks) and 'nan' not in (chs):
                logger.warning(f"For {track['sub'][r]}, {track['ses'][r]} no "
                               "peaks have been provided.")
                if not track['sub'][r] in sub.keys():
                    sub[track['sub'][r]] = [track['ses'][r]]
                else:
                    sub[track['sub'][r]].append(track['ses'][r])
    
    if len(sub) == 0:
        flag = 'approved'
        sub = 'all'
    else:
        flag = 'review'

    return flag
    

def read_manual_peaks(rootpath, sub, ses, chan, adap_bw, logger):
    
    try:
        track = read_tracking_sheet(rootpath, logger)
    except:
        logger.error("Error reading tracking sheet. Check that it isn't open.")
        logger.info("For info how to use adap_bands = 'Manual' in detections, "
                    "refer to documentation:")
        logger.info(" https://seapipe.readthedocs.io/en/latest/index.html")
        logger.info('-' * 10)
        return 'error'

    track = track[track['sub']==sub]
    if len(track.columns) == 0:
        logger.warning(f"Participant not found in column 'sub' in tracking file "
                       "for {sub}, {ses}.")
        return None
    # Search session
    track = track[track['ses']==ses]
    if len(track.columns) == 0:
        logger.warning("Session not found in column 'ses' in tracking file for "
                       f"{sub}, {ses}.")
        return None
    
    # Search channel
    chans = search_chans(track)
    chans = chans.dropna(axis=1, how='all')
    peaks = track.filter(regex='peaks')
    peaks = peaks.dropna(axis=1, how='all')
    
    if len(peaks.columns) == 0:
        logger.warning(f"No spectral peaks found in tracking file for {sub}, {ses}.")
        return None

    chans = chans.to_numpy()[0]
    chans = chans.astype(str)
    chans_all = []
    for cell in chans:
        cell = cell.split(', ')
        for x in cell:
            if ',' in x: 
                x = x.split(',')
                chans_all = chans_all + x
            else:
                chans_all.append(x)
    chans = [x for x in chans_all if not x=='']
    
    
    peaks = peaks.to_numpy()[0]
    peaks = peaks.astype(str)
    peaks_all = []
    for cell in peaks:
        cell = cell.split(', ')
        for x in cell:
            if ',' in x: 
                x = x.split(',')
                peaks_all = peaks_all + x
            else:
                peaks_all.append(x)
    peaks = [float(x) for x in peaks_all if not x=='']
    
    try:
        freq = (peaks[chans.index(chan)] - adap_bw/2, 
                peaks[chans.index(chan)] + adap_bw/2)
    except:
        logger.warning('Inconsistent number of peaks and number of channels '
                       'listed in tracking sheet for {sub}, {ses}. Will use '
                       'Fixed frequency bands instead...')
        freq = None
    
    return freq


def load_adap_bands(tracking, sub, ses, ch, stage, band_limits, adap_bw, logger):
    
    logger.debug(f'Searching for spectral peaks for {sub}, {ses}, {ch}.')
    
    try:
        files = tracking[sub][ses][ch]
    except:
        logger.warning(f'No specparams export file found for {sub}, {ses}, {ch}.')
        return None
    
    files = [x for x in files if stage in x['Stage']]
    files = [x for x in files if band_limits in x['Bandwidth']]

    if len(files) == 0:
        logger.warning(f"No specparams export file found for {sub}, {ses}, {ch}," 
                       f"{stage}, {band_limits}.")
        return None
    elif len(files) > 1:
        logger.warning(f">1 specparams export files found for {sub}, {ses}, {ch}, "
                       f"{stage}, {band_limits} ?")
        return None
    else:
        file = files[0]['File']
    
    
    # Read file and extract peak
    df = read_csv(file)
    df = df.filter(regex='peak')
    df = df.dropna(axis=1, how='all')
    
    if len(df.columns) == 3:
        peak = df.filter(regex='CF').values[0][0]
    elif len(df.columns) == 0: 
        logger.warning(f"No peaks found in export file for {sub}, {ses}, {ch}, "
                       f"{stage}, {band_limits}.")
        return None
    else:
        BW = df.filter(regex='BW')
        maxcol = BW.idxmax(axis='columns')[0].split('_')[1]
        df = df.filter(regex=maxcol)
        peak = df.filter(regex='CF').values[0][0]
            
    freq = (peak - adap_bw/2, 
            peak + adap_bw/2)
    
    return freq
    

def read_inversion(sub, ses, invert, chan, logger):
    
    if type(invert) == type(DataFrame()):
        # Search participant
        chans = invert[invert['sub']==sub]
        if len(chans.columns) == 0:
            logger.warning(f"Participant not found in column 'sub' in tracking "
                           f"file for {sub}, {ses}.")
            return None
        # Search session
        chans = chans[chans['ses']==ses]
        if len(chans.columns) == 0:
            logger.warning(f"Session not found in column 'ses' in tracking file "
                           f"for {sub}, {ses}.")
            return None
        
        # Search channel
        inversion = chans.filter(regex='invert')
        inversion = inversion.dropna(axis=1, how='all')
        chans = search_chans(chans)
        chans = chans.dropna(axis=1, how='all')
        
        if len(inversion.columns) == 0:
            logger.warning(f"No inversion info found in tracking file for "
                           f"{sub}, {ses}.")
            return None

        chans = chans.to_numpy()[0]
        chans = chans.astype(str)
        chans = char.split(chans, sep=', ')
        chans = [x for y in chans for x in y]
        
        inversion = inversion.to_numpy()[0]
        inversion = inversion.astype(str)
        inversion = char.split(inversion, sep=', ')
        inversion = [x for y in inversion for x in y]
        
        if len(inversion) == len(chans):
            inversion = inversion[chans.index(chan)]
            return inversion
        elif len(inversion) == 1:
            return inversion
        else:
            logger.warning(f"Error reading inversion info for {sub}, {ses}, {chan} "
                           "- check documentation for how to provide information "
                           "for inversion:")
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            return None
    

def infer_ref(sub, ses, chan, logger, verbose=0):
    # verbose = 0 (error only)
    # verbose = 1 (warning only)
    # verbose = 2 (debug)
    logger.debug(f"Attempting to infer reference channel for {sub}, {ses} from "
                 "tracking sheet...")
    if isinstance(chan, DataFrame):
        # Search participant
        chans = chan[chan['sub']==sub]
        if len(chans.columns) == 0:
            if verbose>1:
                logger.debug(f"'{sub}' not found in tracking sheet.")
            return None
        # Search session
        chans = chans[chans['ses']==ses]
        if len(chans.columns) == 0:
            if verbose>1:
                logger.debug(f"{ses} for {sub} not found in tracking sheet.")
            return None
        # Search chansets (names in edf)
        oldchans = search_chans(chans)
        oldchans = oldchans.dropna(axis=1, how='all')
        if len(oldchans.columns) == 0:
            if verbose>1:
                logger.debug(f"'Chanset' for {sub} {ses} empty or not found in "
                             "tracking sheet (reference cannot be inferred).")
            return None
        # Search chansets (new names for output)
        newchans = chans.filter(regex='rename')
        newchans = newchans.dropna(axis=1, how='all')
        if len(newchans.columns) == 0:
            if verbose>1:
                logger.debug(f"Column 'newchans' empty or not found in tracking "
                             "sheet, so cannot infer reference channels true name.")
            return None
    else:
        return None
    
    # Infer ref chan name from oldchans:newchans mapping
    if isinstance(oldchans, DataFrame):
        if isinstance(newchans, DataFrame) and len(newchans.columns) != len(oldchans.columns):
            if verbose>0:
                logger.warning(f"There must be the same number of channel sets "
                               "and channel rename sets in tracking file, but "
                               f"for {sub}, {ses}, there were {len(oldchans.columns)} "
                               f"channel sets and {len(newchans.columns)} channel rename sets.")
                if verbose > 1:
                    logger.debug("For info on how to rename channels, refer to documentation:")
                    logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            return None
        
        oldchans = oldchans.to_numpy() 
        oldchans = oldchans.astype(str)
        oldchans = char.split(oldchans[0], sep=', ')
        oldchans = [x for y in oldchans for x in y]
        
        if isinstance(newchans, DataFrame):
            newchans = newchans.to_numpy()
            newchans = newchans.astype(str)
            newchans = char.split(newchans[0], sep=', ')
            newchans = [x for y in newchans for x in y]
        
        if len(oldchans) == len(newchans):
            ref_chan = [newchans[i] for i,chn in enumerate(oldchans) if chn == '_REF']
            if len(ref_chan) < 1:
                if verbose>1:
                    logger.debug(f"No channels named '_REF' in tracking sheet, "
                                 "so cannot infer reference channel.")
                return None
            else:
                ref_chan = ref_chan[0]
        else:
            if verbose>0:
                logger.warning(f"There must be the same number of original channel "
                               "names and new renamed channels in tracking file, "
                               f"but for {sub}, {ses}, there were {len(oldchans)} "
                               f"old channel and {len(newchans)} new channel names.")
                if verbose > 1:
                    logger.debug("For info on how to rename channels, refer to documentation:")
                    logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            return None
    else:
        return None
    
    return ref_chan
    
    
def search_chans(chans):

    chans = chans.filter(regex='chanset')
    chans = chans.filter(regex='^((?!rename).)*$')
    chans = chans.filter(regex='^((?!peaks).)*$')
    chans = chans.filter(regex='^((?!invert).)*$')    
    
    return chans
    
    




# -*- coding: utf-8 -*-
"""


Author: Jordan O'Byrne & formatted by Nathan Cross
"""

from numpy import nan, ones
from os import listdir, mkdir, path
from pandas import DataFrame
from shutil import copy
from wonambi import Dataset
from wonambi.attr import Annotations
from wonambi.detect import match_events
from wonambi.trans import fetch



def event_sync(rec_dir, xml_dir, out_dir, part, visit, cat, 
               evttype_target, evttype_probe, iu_thresh, evttype_tp_target, 
               evttype_fn, chan=None, stage=None, grp=None, rater=None):
    
    
    '''
    This function divides a target event type from XML into events with and without a probe
    '''
    
    
    if path.exists(out_dir):
            print(out_dir + " already exists")
    else:
        mkdir(out_dir)
    
    # loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")       
    
    if visit == 'all': 
        visit = [x for x in listdir(xml_dir + '/' + part[0]) if not '.' in x]
    
    part.sort()
    visit.sort()
    
    print(f"""Merging events... "{evttype_target}" - with - "{evttype_probe}" - into "{evttype_tp_target}"
                    ___
                 , | l | 
                (( | l | ))
                   | l | '
                    \_/
                   /...\--.      _  
                   =====  `-----(_=
                 """)    
    
    # Create output 
    stats_header = ['Recall', 'Precision', 'F1 score', evttype_tp_target, f'{evttype_probe}-', evttype_fn]
    stats = ones((len(part)*len(visit), len(stats_header))) * nan
    
    # Check for channel specification
    if not chan:
        chan = [None]
        
    if not stage:
        stage = [None]
    
    # Concatenation
    cat = list(cat)
    cat[2] = 0
    cat[3] = 0 # force event type non-concatenation (required for analysis)
    cat = tuple(cat)
    
    if cat[1] == 1 and len(stage)>1:
        stage = [stage]
    

    for s, stg in enumerate(stage):
        for ch, channel in enumerate(chan):
            ids = []
            print(f'Channel {channel}; Stage {stg}')
  
            if channel:
                channel = channel + ' (' + grp + ')'

            for i, p in enumerate(part):
                
                
                for v, vis in enumerate(visit):
                    
                    print(f'Subject: {p}, Visit: {vis}')
                    print(f'{channel}')
                    ids.append(f'{p}_{vis}')
                    
                    ## Define files
                    rdir = rec_dir + p + '/' + vis + '/'
                    xdir = xml_dir + p + '/' + vis + '/'
                    edf_file = [x for x in listdir(rdir) if x.endswith('.edf') 
                                or x.endswith('.rec') or x.endswith('.eeg')]
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]            
                    
                    ## Copy annotations file before beginning
                    if not path.exists(out_dir):
                        mkdir(out_dir)
                    if not path.exists(out_dir + p ):
                        mkdir(out_dir + p)
                    if not path.exists(out_dir + p + '/' + vis):
                        mkdir(out_dir + p + '/' + vis)
                        backup = out_dir + p + '/' + vis + '/'
                        backup_file = (f'{backup}{p}_{vis}_spindles.xml')
                        copy(xdir + xml_file[0], backup_file)
                    else:
                        backup = out_dir + p + '/' + vis + '/'
                        backup_file = (f'{backup}{p}_{vis}_spindles.xml')
                    ## Import data
                    dset = Dataset(rdir + edf_file[0])
                    annot = Annotations(backup_file, rater_name=rater)
                    
                    
                    # Get events
                    # target
                    segments = fetch(dset, annot, cat=cat, evt_type=[evttype_target], 
                                     stage=stg, chan_full=[channel], reject_epoch=True, 
                                     reject_artf=['Artefact', 'Arou', 'Arousal'])
                    evt_target_seg = segments.segments
                    print(f'#targets = {len(evt_target_seg)}')
                    for s,seg in enumerate(evt_target_seg):
                        evt_target_seg[s]['start'] = seg['times'][0][0]
                        evt_target_seg[s]['end'] = seg['times'][0][1]
                        evt_target_seg[s]['chan'] = [seg['chan']]
                        evt_target_seg[s]['quality'] = 'Good'
                    
                    # target
                    segments = fetch(dset, annot, cat=cat, evt_type=[evttype_probe], 
                                     stage=stg, chan_full=[channel], reject_epoch=True, 
                                     reject_artf=['Artefact', 'Arou', 'Arousal'])
                    evt_probe_seg = segments.segments
                    print(f'#probes = {len(evt_probe_seg)}')
                    for s,seg in enumerate(evt_probe_seg):
                        evt_probe_seg[s]['start'] = seg['times'][0][0]
                        evt_probe_seg[s]['end'] = seg['times'][0][1]
                        evt_probe_seg[s]['chan'] = [seg['chan']]
                        evt_probe_seg[s]['quality'] = 'Good'
                        
                    
                    # Assess co-occurrence
                    matched = match_events(evt_probe_seg, evt_target_seg, iu_thresh)
         
                    # Write true positive target events to xml
                    matched.to_annot(annot, 'tp_std', evttype_tp_target)
                    
                    if evttype_fn is not None:
                        matched.to_annot(annot, 'fn', evttype_fn)
        


def event_sync_dataset(rec_dir, xml_dir, out_dir, part, visit, cat, outfile_suffix, 
               evttype_target, evttype_probe, iu_thresh, evttype_tp_target, 
               evttype_fn, chan=None, stage=None, grp='eeg', rater=None):
    
    
    '''
        This function accepts 2 event types: {target} and {probe}
        The {target} is split into 2 depending on the presence  of the probe at
        the same time. 
    '''
    
    
    if path.exists(out_dir):
            print(out_dir + " already exists")
    else:
        mkdir(out_dir)
    
    # loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")       
    
    if visit == 'all': 
        visit = [x for x in listdir(xml_dir + '/' + part[0]) if not '.' in x]
    
    part.sort()
    visit.sort()
    
    print(f" Extracting events and creating dataset...")    
    
    # Create output 
    stats_header = ['Recall', 'Precision', 'F1 score', evttype_tp_target, f'{evttype_probe}-', evttype_fn]
    stats = ones((len(part)*len(visit), len(stats_header))) * nan
    
    # Check for channel specification
    if not chan:
        chan = [None]
        
    if not stage:
        stage = [None]
    
    # Concatenation
    cat = list(cat)
    cat[2] = 0
    cat[3] = 0 # force event type non-concatenation (required for analysis)
    cat = tuple(cat)
    
    if cat[1] == 1 and len(stage)>1:
        stage = [stage]
    

    for s, stg in enumerate(stage):
        for ch, channel in enumerate(chan):
            ids = []
            print(f'Channel {channel}; Stage {stg}')
            
            
                
            if channel:
                if not stg:
                    stats_file = out_dir + channel + '_' + outfile_suffix
                elif len(stg)<2:
                    stats_file = out_dir + channel + '_' + stg[0] + '_' + outfile_suffix
                else:
                    if type(stg) is not list:
                        stg = [stg]
                    stages = '_'.join(stg)
                    stats_file = out_dir + channel + '_' + stages + '_' + outfile_suffix
                channel = channel + ' (' + grp + ')'
            else:
                stats_file = out_dir + outfile_suffix 
    
            
            for i, p in enumerate(part):
                
                
                for v, vis in enumerate(visit):
                    
                    print(f'Subject: {p}, Visit: {vis}')
                    print(f'{channel}')
                    ids.append(f'{p}_{vis}')
                    
                    ## Define files
                    rdir = rec_dir + p + '/' + vis + '/'
                    xdir = xml_dir + p + '/' + vis + '/'
                    edf_file = [x for x in listdir(rdir) if x.endswith('.edf') 
                                or x.endswith('.rec') or x.endswith('.eeg')]
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]            
                    
                    ## Import data
                    dset = Dataset(rdir + edf_file[0])
                    annot = Annotations(xdir + xml_file[0], rater_name=rater)
                    
                    # Get events
                    # target
                    segments = fetch(dset, annot, cat=cat, evt_type=[evttype_target], 
                                     stage=stg, chan_full=[channel], reject_epoch=True, 
                                     reject_artf=['Artefact', 'Arou', 'Arousal'])
                    evt_target_seg = segments.segments
                    print(f'#targets = {len(evt_target_seg)}')
                    for s,seg in enumerate(evt_target_seg):
                        evt_target_seg[s]['start'] = seg['times'][0][0]
                        evt_target_seg[s]['end'] = seg['times'][0][1]
                        evt_target_seg[s]['chan'] = [seg['chan']]
                        evt_target_seg[s]['quality'] = 'Good'
                    
                    # target
                    segments = fetch(dset, annot, cat=cat, evt_type=[evttype_probe], 
                                     stage=stg, chan_full=[channel], reject_epoch=True, 
                                     reject_artf=['Artefact', 'Arou', 'Arousal'])
                    evt_probe_seg = segments.segments
                    print(f'#probes = {len(evt_probe_seg)}')
                    for s,seg in enumerate(evt_probe_seg):
                        evt_probe_seg[s]['start'] = seg['times'][0][0]
                        evt_probe_seg[s]['end'] = seg['times'][0][1]
                        evt_probe_seg[s]['chan'] = [seg['chan']]
                        evt_probe_seg[s]['quality'] = 'Good'
                        
                    # Assess co-occurrence
                    matched = match_events(evt_probe_seg, evt_target_seg, iu_thresh)
                    
                    # store stats in master table
                    stats[(i*len(visit))+v, 0] = matched.recall
                    stats[(i*len(visit))+v, 1] = matched.precision
                    stats[(i*len(visit))+v, 2] = matched.f1score
                    stats[(i*len(visit))+v, 3] = matched.n_tp
                    stats[(i*len(visit))+v, 4] = matched.n_fp
                    stats[(i*len(visit))+v, 5] = matched.n_fn
                
            # export stats table
            print(f'Saving {stats_file}')
            df = DataFrame(data=stats, index=ids, columns=stats_header)
            df.to_csv(stats_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 19 14:43:23 2022

@author: nathancross
"""
from safepickle import dump
from os import listdir, mkdir, path
import math
import mne
from numpy import (argsort, asarray, linspace, tile, vstack, zeros)
import shutil
from wonambi import Dataset
from wonambi.trans import fetch
from wonambi.attr import Annotations

'''
    Sleep Undulation Related waveForm (SURF)
    
'''

def marker_to_annot(rec_dir, xml_dir, out_dir, part, visit, rater, chan, 
                    marker_name = None, file_type = 'edf'):
    
    '''
     Reads marker data from EDF directly and exports it into Annotations file 
    '''
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print('')
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")
        print('')
    
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(f'WARNING: input folder missing for Subject {p}, visit {j}, skipping..')
                continue
            else:
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                rec_file = [s for s in listdir(rec_dir + '/' + p + '/' + vis) if ("."+file_type) in s if not s.startswith(".")]
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if x.endswith('.xml') if not x.startswith(".")] 
                
                if len(xml_file) == 0:
                    print(f'WARNING: annotations does not exist for Subject {p}, visit {j} - check this. Skipping..')
                elif len(xml_file) >1:
                    print(f'WARNING: multiple annotations files exist for Subject {p}, visit {j} - check this. Skipping..')
                else:
                    print(f'Annotating embedded edf markers for Subject {p}, Visit {vis}')
                    
                    
                    # Copy annotations file before beginning
                    xdir = xml_dir + '/' + p + '/' + vis + '/' 
                    odir = out_dir + p + '/' + vis + '/'
                    backup_file = (f'{odir}{xml_file[0]}')
                    shutil.copy(xdir + xml_file[0], backup_file)
                    
                    # Open recording and annotations files
                    dset = Dataset(rec_dir + '/' + p + '/' + vis + '/' + rec_file[0]) 
                    annot = Annotations(backup_file, rater_name=rater)

                    # Read markers from edf
                    markers = dset.read_markers()
                    
                    # Filter out markers
                    if marker_name:
                        markers = [x for x in markers if x['name'] in marker_name]
                    
                    # Save markers to output annotations file
                    annot.add_events(markers,chan=chan)

def erp_analysis(rec_dir, xml_dir, out_dir, part, visit, cycle_idx, chan, ref_chan, oREF, 
                 rater, stage,polar, grp_name, cat, evt_type, filt, window,  
                 file_type ='edf'):

    '''
    
    Analyses ERPs
    
    '''
    
    
    ## Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    ## Get number of participants and visits 
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print('')
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")
        print('')
        
    for i, p in enumerate(part):
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        
        
    ## Prepare output array 
    if cycle_idx is None:
        cyc = [1]
    else:
        cyc = cycle_idx
    
    # Axis: sleep cycles
    if cat[0] == 0:
        cyccat = len(cyc)
    else:
        cyccat = 1
          
    # Axis: sleep stages
    if cat[1] == 0:
        stagecat = len(stage)
    else:
        stagecat = 1
    
    # Axis: event types
    if cat[3] == 0:
        evcat = len(evt_type)
    else:
        evcat = 1
    
    # Create output array    
    out = zeros((len(part),len(visit),cyccat,stagecat*evcat), dtype='object')  
           
    
    ## Loop through records
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(f'WARNING: input folder missing for Subject {p}, visit {j}, skipping..')
                continue
            else:
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                rec_file = [s for s in listdir(rec_dir + '/' + p + '/' + vis) if ("."+file_type) in s if not s.startswith(".")]
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if x.endswith('.xml') if not x.startswith(".")] 
                if len(xml_file) == 0:
                    print(f'WARNING: annotations does not exist for Subject {p}, visit {j} - check this. Skipping..')
                elif len(xml_file) >1:
                    print(f'WARNING: multiple annotations files exist for Subject {p}, visit {j} - check this. Skipping..')
                else:
                    # Open recording and annotations files
                    dset = Dataset(rec_dir + '/' + p + '/' + vis + '/' + rec_file[0]) 
                    annot = Annotations(xml_dir + '/' + p + '/' + vis + '/' + xml_file[0], rater_name=rater)
                    
                    # Check polarity of recording
                    if isinstance(polar, list):
                        polarity = polar[i]
                    else:
                        polarity = 'normal'
                       
                    # Get sleep cycles
                    if cycle_idx is not None:
                        all_cycles = annot.get_cycles()
                        scycle = [all_cycles[i - 1] for i in cycle_idx if i <= len(all_cycles)]
                        
                    else:
                        scycle = [None]  

                    chan_full = []
                    for k, ch in enumerate(chan):
                        chan_full.append(ch + ' (' + grp_name + ')')
                            
                    for l, cyc in enumerate(scycle):
                        
                        ## Select and read data
                        print(f'Reading data for {p}, cycle {l+1}')
                        segments = fetch(dset, annot, cat=cat, chan_full=chan_full, 
                                         cycle=[cyc], evt_type=evt_type, stage=stage,
                                         )
                        
                        events = zeros((1,3))
                        event_dict = {}
                        
                        ## Format events
                        for m, seg in enumerate(segments):
                            evtname = seg['name']
                            stg = seg['stage']
                            event_dict[evtname+stg]=m
                            ev = asarray(list(set([x[0]*dset.header['s_freq'] for x in seg['times']])))
                            ev2 = tile(m,(len(ev)))
                            evts = vstack((ev,ev2,ev2)).T
                            events = vstack((events,evts))

                        events = events.astype(int)[1:-1,:]
                        events = events[argsort(events[:, 0])]
                        
                        
                        
                        ## Format data for analysis
                        print('Preparing p, vis for ERP analysis')
                        print(r"""
                              
                                  |
                                  | /\ 
                                  |/  \  _
                              uV2 |   | / \
                                  |   \/   ^-___-__
                                  |________________
                                         (sec)
                              
                              """)
                        
                        achan = chan + ref_chan
                        d = dset.read_data(chan=achan)
                        
                        # Adjust polarity
                        if polarity == 'normal':
                            data = d.data[0]
                        else:
                            data = d.data[0]*-1
                        
                        # Convert data to mne format & filter
                        ch_names = list(d.axis['chan'][0])
                        dig = mne.channels.make_standard_montage('standard_alphabetic')
                        dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
                        info = mne.create_info(ch_names, d.s_freq,verbose=False)
                        mneobj = mne.io.RawArray(data,info,verbose=False)
                        dic = [{x:'eeg' for x in mneobj.ch_names}]
                        mneobj.set_channel_types(dic[0],verbose=False)
                        mneobj.info.set_montage(dig,verbose=False)
                        a = mneobj.pick_types(eeg=True).load_data()
                        a.set_eeg_reference(ref_chan,verbose=False)
                        a.filter(l_freq=filt[0], h_freq=filt[1])    
                        
                        # Save epochs around stimulus event (ERP)
                        epochs = mne.Epochs(a, events, event_id=event_dict, 
                                             tmin=window[0], tmax=window[1], preload=True)

                        # Save individual file
                        with open(out_dir + '/' + p + '/' + vis + '/' + p + '_' + 
                                  '_epochs.p', 'wb') as f:
                             dump(epochs, f)

                        # Save individual epochs to master array
                        for av,avalue in enumerate(event_dict):
                            out[i,j,l,av] = epochs[avalue].average()
        
        # Save master array
        with open(out_dir + f'cat_{cat[0]}{cat[1]}{cat[2]}{cat[3]}_evoked.p', 'wb') as f:
             dump(out, f)
             
    return out
                        
                          
                            
def insert_virtual_markers(rec_dir, xml_dir, out_dir, part, visit, rater, chan, name, reptime,
                           file_type ='edf'):
    
    '''
     Inserts virtual markers into Annotations file 
    '''              
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print('')
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")
        print('')
    
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(f'WARNING: input folder missing for Subject {p}, visit {j}, skipping..')
                continue
            else:
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                #rec_file = [s for s in listdir(rec_dir + '/' + p + '/' + vis) if ("."+file_type) in s if not s.startswith(".")]
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if x.endswith('.xml') if not x.startswith(".")] 
                
                if len(xml_file) == 0:
                    print(f'WARNING: annotations does not exist for Subject {p}, visit {j} - check this. Skipping..')
                elif len(xml_file) >1:
                    print(f'WARNING: multiple annotations files exist for Subject {p}, visit {j} - check this. Skipping..')
                else:
                    print(f'Creating virtual markers for Subject {p}, Visit {vis}')
                    
                    
                    # Copy annotations file before beginning
                    xdir = xml_dir + '/' + p + '/' + vis + '/' 
                    odir = out_dir + p + '/' + vis + '/'
                    backup_file = (f'{odir}{xml_file[0]}')
                    shutil.copy(xdir + xml_file[0], backup_file)
                    
                    # Open recording and annotations files
                    #dset = Dataset(rec_dir + '/' + p + '/' + vis + '/' + rec_file[0]) 
                    annot = Annotations(backup_file, rater_name=rater)  
                         
                    # Get start time of first sleep stage
                    epochs = annot.get_epochs()   
                    first_epoch = [x for x in epochs if x['stage'] in ['NREM1','NREM2','NREM3','REM']][0]
                    onset = int(first_epoch['start'])
                    
                    # Get end time of last sleep stage
                    last_epoch = [x for x in epochs if x['stage'] in ['NREM1','NREM2','NREM3','REM']][-1]
                    offset = int(last_epoch['end'])
                    
                    
                    # Create number of events (based on repition time)
                    num_evts = (offset-onset)/reptime
                    
                    # Create list of start times for events
                    if num_evts.is_integer():
                        starts = linspace(onset, offset, int(num_evts))
                    else:
                        num_evts = math.ceil(num_evts)
                        offset = int(onset + (num_evts*reptime))
                        starts = linspace(onset, offset, num_evts)
                    
                    # Create list of evts  
                    evts = []
                    for x, st in enumerate(starts):
                        evts.append({'name': name,
                         'start': st,
                         'end': st + 0.5,
                         'chan': chan,
                         'stage': '',
                         'quality': 'Good'})

                    # Add events to annotations file
                    annot.add_events(evts)
                    
                    
                    
def marker_to_annotAP(rec_dir, xml_dir, out_dir, part, visit, rater, chan, marker_name=None, file_type='edf'):
    
    '''
     Reads marker data from EDF directly and exports it into Annotations file 
    '''
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Loop through records
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print('')
        print("ERROR: 'part' must either be an array of subject ids or = 'all' ")
        print('')
    
    for i, p in enumerate(part):
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        for j, vis in enumerate(visit): 
            if not path.exists(xml_dir + '/' + p + '/' + vis + '/'):
                print(f'WARNING: input folder missing for Subject {p}, visit {j}, skipping..')
                continue
            else:
                if not path.exists(out_dir + '/' + p + '/' + vis):
                    mkdir(out_dir + '/' + p + '/' + vis)
                rec_file = [s for s in listdir(rec_dir + '/' + p + '/' + vis) if ("."+file_type) in s if not s.startswith(".")]
                xml_file = [x for x in listdir(xml_dir + '/' + p + '/' + vis) if x.endswith('.xml') if not x.startswith(".")] 
                
                if len(xml_file) == 0:
                    print(f'WARNING: annotations does not exist for Subject {p}, visit {j} - check this. Skipping..')
                elif len(xml_file) >1:
                    print(f'WARNING: multiple annotations files exist for Subject {p}, visit {j} - check this. Skipping..')
                else:
                    print(f'Annotating embedded edf markers for Subject {p}, Visit {vis}')
                    
                    
                    # Copy annotations file before beginning
                    xdir = xml_dir + '/' + p + '/' + vis + '/' 
                    odir = out_dir + p + '/' + vis + '/'
                    backup_file = (f'{odir}{xml_file[0]}')
                    shutil.copy(xdir + xml_file[0], backup_file)
                    
                    # Open recording and annotations files
                    dset = Dataset(rec_dir + '/' + p + '/' + vis + '/' + rec_file[0]) 
                    annot = Annotations(backup_file, rater_name=rater)

                    # Read markers from edf
                    markers = dset.read_markers()
                    
                    # Filter out markers
                    if marker_name:
                        markers = [x for x in markers if x['name'] in marker_name]
                    
                    # Select every 2nd element    
                    markers = markers[1::2]
                    
                    # Save markers to output annotations file
                    annot.add_events(markers,chan=chan)                    
                                  
                                    
                                    
                                    
                                    
                                    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  7 15:47:20 2025

@author: ncro8394
"""

from numpy import (angle, arange, array, average, concatenate, degrees, diff, exp,
                   histogram,  inf, isnan, linspace, max, min, nan, nanmean,
                   nanmedian, nanstd, pi, random, sort, vstack, where, zeros)
from pandas import DataFrame, Series
from collections import Counter, defaultdict
from copy import deepcopy
from os import listdir, mkdir, path
import pywt
import shutil
from scipy.optimize import curve_fit
from scipy.interpolate import interp1d
from scipy.signal import (butter, firwin, filtfilt, find_peaks, detrend, 
                          hilbert, morlet2, periodogram, resample, welch) 
from sklearn.cluster import DBSCAN
from wonambi import Dataset
from wonambi.attr import Annotations
from wonambi.detect import consensus, DetectSpindle
from wonambi.trans import fetch
from ..utils.logs import create_logger, create_logger_outfile
from ..utils.load import (load_channels, load_adap_bands, load_sessions, 
                          rename_channels, read_manual_peaks)
from ..utils.misc import remove_duplicate_evts, merge_epochs


class clam:

        def __init__(self, rootpath, rec_dir, xml_dir, out_dir, chan, 
                     grp_name, stage, rater = None, subs = 'all', 
                     sessions = 'all', tracking = None):
            
            self.rootpath = rootpath
            self.rec_dir = rec_dir
            self.xml_dir = xml_dir
            self.out_dir = out_dir
            self.chan = chan
            self.grp_name = grp_name
            self.stage = stage
            self.rater = rater
            
            self.subs = subs
            self.sessions = sessions
            
            if tracking == None:
                tracking = {'cluster':{}}
            self.tracking = tracking

        def clustering(self, evt_type = 'spindle', 
                             freq_bands = {'SWA': (0.5, 4), 'Sigma': (10, 15)},
                             filetype = '.edf', grp_name = 'eeg',
                             logger = create_logger('Event clustering')):
            
            
            
            ### 0.a Set up logging
            tracking = self.tracking
            flag = 0
            
            logger.info('')            
            logger.debug(rf"""Commencing clustering and fluctuations pipeline...
       
                                  __.--:__:--.__
                             _.--°      |       °--._
                            |    \      |      /     ;
                            \     \     |     /     /
                             `\    \    |    /    /'
                               `\   \   |   /   /'
                                 `\  \  |  /  /'
                                _.-`\ \ | / /'-._
                               (_____`\\|//'_____)      
                                       `-'
                  
                    Clustering and Low-frequency Activity Modulations 
                    (C.L.A.M)   
                    
                    Event type: {evt_type}
                                                """,)
            
            
            ### 1. First we check the directories
            # a. Check for output folder, if doesn't exist, create
            if path.exists(self.out_dir):
                    logger.debug("Output directory: " + self.out_dir + " exists")
            else:
                mkdir(self.out_dir)
            
            # b. Check input list
            subs = self.subs
            if isinstance(subs, list):
                None
            elif subs == 'all':
                    subs = listdir(self.rec_dir)
                    subs = [p for p in subs if not '.' in p]
            else:
                logger.error("'subs' must either be an array of subject ids or = 'all' ")       
            
            ### 2. Begin loop through dataset
           
            # a. Begin loop through participants
            subs.sort()
            for i, sub in enumerate(subs):
                if not sub in self.tracking['cluster'].keys():
                    self.tracking['cluster'][sub] = {}
                
                # b. Begin loop through sessions
                flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                         logger, verbose=2)   
                for v, ses in enumerate(sessions):
                    logger.info('')
                    logger.debug(f'Commencing {sub}, {ses}')
                    if not ses in self.tracking['cluster'][sub].keys():
                        self.tracking['cluster'][sub][ses] = {} 
        
                    
                    ## c. Load recording
                    rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
                    try:
                        edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                        dset = Dataset(rdir + edf_file[0])
                    except:
                        logger.warning(f' No input {filetype} file in {rdir}')
                        flag+=1
                        break
                    
                    ## d. Load annotations
                    xdir = self.xml_dir + '/' + sub + '/' + ses + '/'
                    if not path.exists(xdir):
                        logger.warning(f'{evt_type} has not been detected for'
                                       f'{sub}, {ses}. Skipping..')
                        flag += 1
                        continue
                    xml_file = [x for x in listdir(xdir) if x.endswith('.xml')][0]
                    # Copy annotations file before beginning
                    if not path.exists(self.out_dir):
                        mkdir(self.out_dir)
                    if not path.exists(self.out_dir + '/' + sub):
                        mkdir(self.out_dir + '/' + sub)
                    if not path.exists(self.out_dir + '/' + sub + '/' + ses):
                        mkdir(self.out_dir + '/' + sub + '/' + ses)
                    outpath = self.out_dir + '/' + sub + '/' + ses + '/'
                    backup_file = (f'{outpath}{sub}_{ses}.xml')
                    if not path.exists(backup_file):
                        shutil.copy(xdir + xml_file, backup_file)

                    # Read annotations file
                    annot = Annotations(backup_file, rater_name=self.rater)
                    
                    ## e. Channel setup 
                    pflag = deepcopy(flag)
                    flag, chanset = load_channels(sub, ses, self.chan, self.chan,
                                                  flag, logger)
                    if flag - pflag > 0:
                        logger.warning(f'Skipping {sub}, {ses}...')
                        flag += 1
                        continue
                    
                    newchans = rename_channels(sub, ses, self.chan, logger)
                    
                    for c, ch in enumerate(chanset):
                        
                        # 5.b Rename channel for output file (if required)
                        if newchans:
                            fnamechan = newchans[ch]
                        else:
                            fnamechan = ch
                            
                        stagename = '-'.join(self.stage)
                        
                        
            
                        # ---- Analys event clustering ----
                        logger.debug(f'Analysing cluster metrics for {evt_type}, {ch}')
                        stats = cluster_metrics(annot, ch, evt_type, 
                                                   self.stage, grp_name)
                        
                        # Save stats to file
                        if not isinstance(stats, dict):
                            flag += 1
                            continue
                        else:
                            df = DataFrame([stats])
                            df.insert(0, 'sub', sub)  
                            df.insert(1, 'ses', ses)   
                            logger.debug(f'Saving cluster metrics: {evt_type}')
                            df.to_csv(f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{evt_type}_clustering.csv')
                            
                            
                            
                        # ---- Analyse low-frequency fluctuations ----
                        logger.debug('Analysing low frequency fluctuations for '
                                     f"{', '.join([band for band in freq_bands])} "
                                     f'in channel: {fnamechan}: {chanset[ch]}')
                        segs = fetch(dset, annot, cat = (0,1,1,1), 
                                     stage = self.stage,
                                      reject_artf = True)
                        segs.read_data(chan = [ch], ref_chan = chanset[ch])
                        baseline_data = segs[0]['data'].data[0][0]
                        sampling_rate = segs[0]['data'].s_freq
                        baseline_data = baseline_data[0:int(sampling_rate*100)]
                        
                        segs = fetch(dset, annot, cat = (0,0,0,0), 
                                     stage = self.stage,
                                      reject_artf = True)
                        segs.read_data(chan = [ch], ref_chan = chanset[ch])
                        
                        psd = {band:[] for band in freq_bands}
                        dur = {band:[] for band in freq_bands}
                        phases = {band:[] for band in freq_bands}
                        trace = {band:zeros((2,1)) for band in freq_bands}
                        results = None
                        for i, seg in enumerate(segs):
                            signal_data = seg['data'].data[0][0]
                            if len(signal_data)/sampling_rate > 300:
                                results = extract_infraslow_spectral_profile(signal_data, 
                                                                             sampling_rate,
                                                                             baseline_data,
                                                                             band_defs=freq_bands)
                                
                                for band in freq_bands:
                                    psd[band].append(results[band]['psd'])
                                    dur[band].append(results[band]['duration'])

                                    ## ---- PHASE ANGLE OF SPINDLE OCCURENCES ----
                                    # Get spindle events
                                    start_seg = seg['data'].axis['time'][0].min()
                                    end_seg = seg['data'].axis['time'][0].max()
                                    events = [x for x in annot.get_events(evt_type, chan = f'{ch} ({grp_name})') if 
                                              x['start'] > start_seg and x['end'] < end_seg]
                                    evts_mid = [ (x['start'] - start_seg) + 
                                                ((x['end'] - start_seg) - (x['start'] - start_seg))/2
                                                for x in events]
                                    #evts_idx = round(array(evts_mid) * resample_rate).astype(int)
                                    evts_idx = [int(x*2) for x in evts_mid]
                                    
                                    # Calculate phase angle of spindle mid-point
                                    sigma_trace = results[band]['smoothed_power_trace']
                                    sigma_trace = sigma_trace - nanmean(sigma_trace)
                                    
                                    smoothed_signal = Series(sigma_trace).rolling(window=5, 
                                                                                center=True,
                                                                                min_periods=1).mean().to_numpy()
                                    
                                    smoothed_signal = resample(smoothed_signal, 
                                                               len(sigma_trace))
                                    analytic_signal = hilbert(smoothed_signal)
                                    instantaneous_phase = angle(analytic_signal)
                                    event_phases = instantaneous_phase[evts_idx]
                                    phases[band].extend(event_phases.tolist())
                                    
                                    # Save filtered trace
                                    l = linspace(int(start_seg),
                                                 int(end_seg), 
                                                 int(end_seg)- int(start_seg)+1)
                                    
                                    signal_1s = smoothed_signal[::2]
                                    
                                    if len(signal_1s) > len(l):
                                        signal_1s = signal_1s[:len(l)]
                                    elif len(signal_1s) < len(l):
                                        l = l[:len(signal_1s)]
                                    trace[band] = concatenate((trace[band], 
                                                               vstack((l, signal_1s))),
                                                              axis = 1)
                                    
                                # End the analysis (only analyse the first segment >300 sec)    
                                break
                                
      
                                 
                        # Peak of averaged power spectrum
                        if results:
                            stats = {}
                            for band in freq_bands:
                                arrays = array(psd[band])           
                                weights = array(dur[band])
                                weighted_avg = average(arrays, 
                                                       axis=0, 
                                                       weights=weights) if len(weights)>0 else average(arrays, axis=0)
                                mean_psd = weighted_avg / nanmean(weighted_avg)
                                             
                                # Gaussian fit 
                                freqs_low = results[band]['infraslow_freqs']
                                (peak_freq, 
                                 sigma, 
                                 avg_peak_power) = fit_gaussian(
                                                                freqs_low[(freqs_low > 0.01) & (freqs_low < 0.05)],
                                                                mean_psd[(freqs_low > 0.01) & (freqs_low < 0.05)]
                                                                )
                                logger.debug(f"Chan: {fnamechan}: {chanset[ch]} - "
                                             f"{band} infraslow peak: {round(peak_freq,3)}; "
                                             f"Peak power: {round(avg_peak_power,3)}")
                                
                                stats[f'{band}_peak_freq'] = round(peak_freq,3)
                                stats[f'{band}_avg_peak_power'] = round(avg_peak_power,3)
                                
                                # Histogram of events
                                bins = linspace(-pi, pi, 19)
                                hist, bin_edges = histogram(phases[band], bins=bins)
                                bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
                                hist_df = DataFrame(hist).transpose()
                                hist_df.columns = [int(round(degrees(x))) for x in bin_centers]
                                hist_df.to_csv(f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{evt_type}_{band}_coupling.csv')
                                
                                # PSD of LowFreq-Fluctuation
                                psd_df = DataFrame(mean_psd).transpose()
                                psd_df.columns = [f'{round(f,3)} Hz' for f in freqs_low]
                                psd_df.to_csv(f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{band}_fluctuations_psd.csv')
                                
                                # Filtered trace
                                ts_df = DataFrame([trace[band][1]], 
                                                  columns = trace[band][0])
                                ts_df.to_csv(f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{band}_fluctuations_timeseries.csv')
    
                                
                            # Save stats to file
                            df = DataFrame([stats])
                            df.insert(0, 'sub', sub)  
                            df.insert(1, 'ses', ses)   
                            bandnames = '-'.join([band for band in freq_bands])
                            df.to_csv(f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_{bandnames}_fluctuation_stats.csv')
                        else:
                            logger.warning(f'No valid segments found for {sub}. {ses}. Skipping...')
                            flag +=1

            ### 3. Check completion status and print
            if flag == 0:
                logger.debug('Event clustering analysis finished without error.')  
            else:
                logger.warning(f'Event clustering analysis finished with {flag} WARNINGS. See log for details.')
            
            #self.tracking = tracking   ## TO UPDATE - FIX TRACKING
            
            return 
            
            
            
def cluster_metrics(annot, chan, evt_name = 'spindle', stage = None, 
                       grp_name = 'eeg', 
                       logger = create_logger('Event clustering')):
    
    
    """
    Compute temporal clustering metrics for detected sleep events (e.g., spindles).

    This function quantifies the degree to which events (such as spindles) are temporally clustered 
    within a recording, including the coefficient of variation (CV), variance-to-mean ratio (VMR), 
    and DBSCAN-based cluster statistics. It also performs permutation-based null testing to evaluate 
    statistical significance of clustering.

    Parameters
    ----------
    annot : wonambi.Annotations
        Wonambi annotations object containing event and staging information.
    chan : str
        Name of the EEG channel to extract events from.
    evt_name : str, optional
        Name of the event type to analyze (default is 'spindle').
    stage : list of str or None, optional
        List of sleep stages (e.g., ['NREM2', 'NREM3']) to restrict event analysis to.
        If None, all stages are used.
    grp_name : str, optional
        Name of the signal group in the annotations (default is 'eeg').
    logger : logging.Logger, optional
        Logger instance for progress and warning messages.

    Returns
    -------
    stats : dict
        Dictionary of summary statistics describing temporal clustering of events:

        - 'cv' : float
            Coefficient of variation of inter-event intervals.
        - 'cv_p_value' : float
            P-value from permutation test for CV.
        - 'VMR' : float
            Variance-to-mean ratio of rolling event counts.
        - 'VMR_p_value' : float
            P-value from Monte Carlo test for VMR.
        - 'num_clusters_n' : int
            Number of DBSCAN-detected clusters.
        - 'average_cluster_size_n' : float
            Average number of events per cluster.
        - 'max_cluster_size_n' : int
            Size of the largest cluster.
        - 'min_cluster_size_n' : int
            Size of the smallest cluster.
        - 'clusters_per_60s' : float
            Number of clusters per minute of analyzed time.
        - 'ave_within_cluster_freq' : float
            Average frequency (Hz) of events within clusters.
        - 'mean_iei_s' : float
            Mean inter-event interval (in seconds) within clusters.
        - 'median_iei_s' : float
            Median inter-event interval (in seconds) within clusters.
        - 'min_interval_s' : float
            Minimum interval (in seconds) between clusters.
        - 'mean_interval_s' : float
            Mean interval (in seconds) between clusters.
        - 'max_interval_s' : float
            Maximum interval (in seconds) between clusters.
        - 'min_cluster_duration_s' : float
            Shortest cluster duration (in seconds).
        - 'avg_cluster_duration_s' : float
            Average cluster duration (in seconds).
        - 'max_cluster_duration_s' : float
            Longest cluster duration (in seconds).

    Notes
    -----
    - Events must be pre-detected and stored in the provided annotations object.
    - Clustering uses DBSCAN with fixed parameters (eps=20, min_samples=3).
    - If fewer than 3 events are found, clustering statistics will not be returned.
    """


    # Stages
    stage_key = {'Wake':0,
                 'NREM1':1,
                 'NREM2':2,
                 'NREM3':3,
                 'REM':5,
                 'Artefact':0,
                 'Undefined':0,
                 'Unknown':0}
    
    epochs = annot.get_epochs()
    leng = epochs[-1]['end']
    
    hypno = zeros(leng)
    
    for epoch in epochs:
        start = int(epoch['start'])
        end = int(epoch['end'])
        
        hypno[start:end] = stage_key[epoch['stage']]

    merged = merge_epochs(epochs)
    if stage:
        merged = [x for x in merged if x['stage'] in stage]
    
    # Get Spindles
    chan_full = chan + ' (' + grp_name + ')'
    events = annot.get_events(evt_name, chan = chan_full, stage = stage)
    if len(events) == 0:
        logger.warning(f'No {evt_name} events found in Annotations file.')
        return 'error'
    evt_array = zeros(leng)
    
    for ev in events:
        start = int(ev['start'])
        end = int(ev['end'])
        evt_array[start:end] = 1
    
    evt_array_trim = array([0])
    for epoch in merged:
            segment = evt_array[epoch['start']:epoch['end']]
            evt_array_trim = concatenate((evt_array_trim,segment))
    
    num_permutations = 1000
    # Calculate temporal Coefficient of Variation
    event_times = where(evt_array_trim == 1)[0]  # Indices of events
    inter_event_times = diff(event_times)
    
    if len(inter_event_times) > 1:
        cv = nanstd(inter_event_times) / nanmean(inter_event_times)
    
        # Permutation test: Shuffle event locations to create a null distribution
        cv_null = []
        event_indices = sort(random.choice(arange(100), 
                                                 size=10, 
                                                 replace=False))  # Random non-adjacent events
        for _ in range(num_permutations):
            shuffled_events = random.choice(arange(100), 
                                               size=len(event_indices),
                                               replace=False)
            shuffled_iet = diff(sort(shuffled_events))
            cv_null.append(nanstd(shuffled_iet) / nanmean(shuffled_iet))
        
        # Compute p-value: fraction of null CVs greater than observed CV
        cv_p_value = sum(array(cv_null) >= cv) / num_permutations
        #print(f"Observed CV of inter-event times: {cv:.2f}, p-value: {cv_p_value:.3f}")
    
    
    # Calculate the observed VMR (Variance-to-Mean Ratio)
    window_size = 30
    rolling_counts = Series(evt_array_trim).rolling(window=window_size).sum()
    rolling_counts = rolling_counts[::5]
    observed_vmr = rolling_counts.var() / rolling_counts.mean()
    
    # Monte Carlo Simulation
    vmr_null = []
    for _ in range(num_permutations):
        shuffled_series = zeros(100)
        shuffled_events = random.choice(arange(100), size=len(event_indices), replace=False)
        shuffled_series[shuffled_events] = 1
        shuffled_counts = Series(shuffled_series).rolling(window=window_size).sum()
        shuffled_counts = shuffled_counts[::5]
        vmr_null.append(shuffled_counts.var() / shuffled_counts.mean())
    
    # Compute p-value
    vmr_p_value = sum(array(vmr_null) >= observed_vmr) / num_permutations
    #print(f"Observed VMR: {observed_vmr:.2f}, p-value: {vmr_p_value:.3f}")
    

    #### Find clusters and get parameters    
    if len(event_times) >= 3:  # DBSCAN needs at least min_samples
        db = DBSCAN(eps=20, min_samples=3, metric="euclidean").fit(event_times.reshape(-1, 1))
        labels = db.labels_
        
        # ---- Get unique clusters (excluding noise: label -1) ----
        valid_cluster_labels = set(labels) - {-1}
        num_clusters = len(valid_cluster_labels)
        
        # ---- Average cluster size ----
        cluster_counts = Counter(labels)
        if -1 in cluster_counts:  # Remove noise label (-1)
            del cluster_counts[-1]
        average_cluster_size = nanmean(list(cluster_counts.values()))

        # ---- Clusters per unit time (e.g., per 30 seconds) ----
        time_range = sum([x['end'] - x['start'] for x in merged])
        clusters_per_60s = num_clusters / (time_range / 60)
        
        
        # ---- Group events by cluster ----
        clusters = defaultdict(list)
        for t, label in zip(event_times.flatten(), labels):
            if label != -1:  # skip noise
                clusters[label].append(t)
                
        # ---- Min/Max cluster size ----
        cluster_sizes = [len(times) for times in clusters.values() if len(times) > 0]
        if len(cluster_sizes) > 0:
            max_cluster_size = max(cluster_sizes)
            min_cluster_size = min(cluster_sizes)   
        else:
            max_cluster_size = nan
            min_cluster_size = nan
        
        
        # ---- Calculate within-cluster frequency ----
        cluster_freqs = {}
        for label, times in clusters.items():
            times = sorted(times)
            duration = times[-1] - times[0]
            n_events = len(times)
        
            if duration > 0:  # avoid division by zero
                freq = n_events / duration  # events per second
                cluster_freqs[label] = freq
            else:
                cluster_freqs[label] = nan  # maybe a single event or nearly simultaneous
       
        # Summary stats 
        valid_freqs = [f for f in cluster_freqs.values() if not isnan(f)]
        average_within_cluster_freq = nanmean(valid_freqs)
        
        
        # ---- Calculate IEIs per cluster ----
        cluster_ieis = {}  # store IEIs per cluster
        all_ieis = []      # collect all IEIs if you want overall stats
        
        for label, times in clusters.items():
            sorted_times = sorted(times)
            
            if len(sorted_times) >= 2:
                ieis = diff(sorted_times)
                cluster_ieis[label] = ieis
                all_ieis.extend(ieis)
            else:
                cluster_ieis[label] = array([])  # cluster with only one event
        
        # Summary stats
        mean_iei = nanmean(all_ieis)
        median_iei = nanmedian(all_ieis)
        
        
        # ---- Between cluster intervals & cluster duration ----
        # Build (start, end) pairs for each cluster
        cluster_bounds = []
        cluster_durations = []
        for label, times in clusters.items():
            if len(times) > 1:
                start = min(times)
                end = max(times)
                cluster_bounds.append((start, end))
                duration = max(times) - min(times)
                cluster_durations.append(duration)

        cluster_bounds.sort() # Sort clusters by start time
        
        # Compute between-cluster intervals
        between_cluster_intervals = []
        for i in range(len(cluster_bounds) - 1):
            current_end = cluster_bounds[i][1]
            next_start = cluster_bounds[i+1][0]
            interval = next_start - current_end
            between_cluster_intervals.append(interval)
        
        # Cluster duration
        if len(cluster_durations) > 0:
            avg_cluster_duration = nanmean(cluster_durations)
            min_cluster_duration = min(cluster_durations)
            max_cluster_duration = max(cluster_durations)
        else:
            avg_cluster_duration = min_cluster_duration = max_cluster_duration = nan
        
        # Between cluster intervals
        if len(between_cluster_intervals) > 0:
            min_interval = min(between_cluster_intervals)
            mean_interval = nanmean(between_cluster_intervals)
            max_interval = max(between_cluster_intervals)
        else:
            min_interval = mean_interval = max_interval = nan

        stats = {'cv': round(cv, 2),
                 'cv_p_value': round(cv_p_value, 3),
                 'VMR': round(observed_vmr,2),
                 'VMR_p_value': round(vmr_p_value, 3),
                 'num_clusters_n': num_clusters,
                 'average_cluster_size_n': round(average_cluster_size,2),
                 'max_cluster_size_n': round(max_cluster_size,2),
                 'min_cluster_size_n': round(min_cluster_size,2),
                 'clusters_per_60s': round(clusters_per_60s,2),
                 'ave_within_cluster_freq': round(average_within_cluster_freq,4),
                 'mean_iei_s': round(mean_iei, 2),
                 'median_iei_s': round(median_iei, 2),
                 'min_interval_s': round(min_interval, 2),
                 'mean_interval_s': round(mean_interval, 2),
                 'max_interval_s': round(max_interval, 2),
                 'min_cluster_duration_s': round(min_cluster_duration, 2),
                 'avg_cluster_duration_s': round(avg_cluster_duration, 2),
                 'max_cluster_duration_s': round(max_cluster_duration, 2),
                }
        
    else:
        logger.warning("Not enough events in segment to compute clustering.")
        stats = 'error'
        
    return stats            
            


def extract_infraslow_spectral_profile(eeg_signal, fs, baseline_data,
                                        wavelet='cmor4.0-1.0',
                                        band_defs={'SWA': (0.5, 4), 'Sigma': (10, 15)},
                                        smoothing_window_sec = 4,
                                        infraslow_freq_range = (0.001, 0.12),
                                        infraslow_step = 0.001,
                                        downsample_step_sec = 0.5):
    
    """
    Extract infraslow power fluctuations in EEG band-specific power using a two-stage wavelet decomposition.

    This function computes smoothed and normalized power time courses for specified frequency bands
    (e.g., SWA, sigma), then applies a second wavelet transform to quantify infraslow fluctuations
    in these band power signals.
    
    This function is based upon the paper: 
        Lecci et al. (2017) - Science Advances;  
        'Coordinated infraslow neural and cardiac oscillations mark fragility 
        and offline periods in mammalian sleep.'

    Parameters
    ----------
    eeg_signal : array_like
        1D array of EEG signal values (time series).
    fs : float
        Sampling frequency of the EEG signal in Hz.
    baseline_data : array_like
        EEG data (typically from a baseline period) used to compute average band power for normalization.
    wavelet : str, optional
        Wavelet name used for CWT (default is 'cmor4.0-1.0').
    band_defs : dict, optional
        Dictionary defining frequency bands to analyze, with names as keys and (fmin, fmax) tuples as values.
    smoothing_window_sec : float, optional
        Duration of the moving average smoothing window in seconds (default is 4).
    infraslow_freq_range : tuple, optional
        Range of infraslow frequencies (in Hz) for the second-stage wavelet transform (default is (0.001, 0.12)).
    infraslow_step : float, optional
        Step size (Hz) for infraslow frequency resolution (default is 0.001).
    downsample_step_sec : float, optional
        Time resolution (seconds) for downsampling the smoothed power traces before infraslow analysis (default is 0.5).

    Returns
    -------
    results : dict
        Dictionary with a key for each band in `band_defs`. Each value is a dictionary with:
        
        - 'infraslow_freqs' : ndarray
            Frequencies used in the second-stage (infraslow) wavelet transform.
        - 'psd' : ndarray
            Average power spectral density across time points for the band-specific infraslow transform.
        - 'smoothed_power_trace' : ndarray
            The downsampled, smoothed, and normalized power time series used for the infraslow transform.
        - 'duration' : int
            Number of original EEG samples (i.e., length of `eeg_signal`).
    """
    
    dt = 1 / fs
    n = len(eeg_signal)

    # Time-frequency decomposition (0.5–24 Hz)
    freqs_high = arange(0.5, 24.1, 0.2)
    scales_high = pywt.central_frequency(wavelet) / (freqs_high * dt)
    coeffs, freqs_used = pywt.cwt(eeg_signal, scales_high, wavelet, sampling_period=dt)
    power = abs(coeffs) ** 2

    results = {}
    
    # Calculate PSD on entire segment, for normalisation.
    f_fft, pxx = welch(baseline_data, fs=fs, nperseg=fs*4)
    

    for band_name, (fmin, fmax) in band_defs.items():
        
        # Band power time series
        band_idx = (freqs_used >= fmin) & (freqs_used <= fmax)
        band_power = power[band_idx].mean(axis=0)

        # Moving average smoothing
        win_samples = int(smoothing_window_sec * fs)
        smooth = Series(band_power).rolling(win_samples, center=True, min_periods=1).mean().to_numpy()

        # Normalize to mean power
        idx_band = (f_fft >= fmin) & (f_fft <= fmax)
        band_baseline = nanmean(pxx[idx_band])
        normalized = smooth / band_baseline

        # Downsample for infraslow resolution
        step = int(downsample_step_sec * fs)
        smooth_ds = normalized[::step]
        dt_ds = downsample_step_sec

        # Infraslow wavelet transform (e.g. 0.001–0.12 Hz)
        freqs_low = arange(infraslow_freq_range[0], infraslow_freq_range[1], infraslow_step)
        scales_low = pywt.central_frequency(wavelet) / (freqs_low * dt_ds)
        coeffs_low, _ = pywt.cwt(smooth_ds, scales_low, wavelet, sampling_period=dt_ds)
        power_low = abs(coeffs_low) ** 2
        mean_power = power_low.mean(axis=1)

        results[band_name] = {
                              'infraslow_freqs': freqs_low,
                              'psd': mean_power,
                              'smoothed_power_trace': smooth_ds,
                              'duration': n
                             }

    return results


def gaussian(x, a, mu, sigma):
    return a * exp(-((x - mu) ** 2) / (2 * sigma ** 2))

def fit_gaussian(freqs_low, psd_vals):
    """
    Fit a Gaussian curve to a 1D power spectrum and compute peak characteristics.

    Parameters
    ----------
    freqs_low : array_like
        1D array of frequency values (e.g., in the infraslow range).
    psd_vals : array_like
        1D array of power spectral density (PSD) values corresponding to `freqs_low`.

    Returns
    -------
    peak_freq : float
        Estimated peak frequency (mu) of the fitted Gaussian [Hz].
    sigma : float
        Standard deviation (spread) of the fitted Gaussian [Hz].
    avg_peak_power : float
        Average PSD value in the frequency window centered at the peak frequency
        and spanning ±0.5 * sigma, computed via linear interpolation.
        Returns NaN for all outputs if the fit fails.
    """
    
    try:
        # Example: constrain mu to [0.01, 0.05], and sigma > 0
        bounds = ([0, 0.01, 1e-4], [inf, 0.05, 0.1])  # [a, mu, sigma]
        
        # Fit Gaussian
        popt, _ = curve_fit(gaussian, freqs_low, psd_vals, 
                            p0 = [1.0, 0.02, 0.01],
                            bounds = bounds)
        a, peak_freq, sigma = popt
        peak_band = (peak_freq - 0.5 * sigma, peak_freq + 0.5 * sigma)
        
        # Clip window to valid frequency range
        peak_band = (
                    max((peak_band[0], freqs_low.min())),
                    min((peak_band[1], freqs_low.max()))
                    )
        
        # Interpolate PSD for smoother power estimate in peak window
        interp_fn = interp1d(freqs_low, psd_vals, 
                             kind='linear', 
                             bounds_error=False, 
                             fill_value='extrapolate')
        fine_freqs = linspace(*peak_band, 100)
        fine_power = interp_fn(fine_freqs)
        
        avg_peak_power = nanmean(fine_power)
        
    except RuntimeError:
        peak_freq, sigma, avg_peak_power = nan, nan, nan

    return peak_freq, sigma, avg_peak_power
    
            


# -*- coding: utf-8 -*-
"""
Created on Mon May 13 13:10:09 2019

CFC Functions

@author: jordan
"""

from numpy import (any, arange, arctan2, argmin, around, array, asarray, 
                   atleast_2d, ceil, concatenate, cumsum, dot, empty, exp, floor, genfromtxt, 
                   hstack, hypot, identity, linalg, linspace, log, logical_and, 
                   max, mean, minimum, multiply, nan, nanmean, nanstd, ones, 
                   pi, power, prod, reshape, roll, sin, sqrt, squeeze, std, sum, 
                   transpose, vstack, where, zeros)
from os import listdir, mkdir, path
from scipy.stats import f as fdist
from scipy.special import lpn
import matplotlib.pyplot as plt
from pingouin import circ_r   
from safepickle import load
import warnings


def circ_wwtest(alpha1, alpha2, w1, w2, warnings=True):
    """Parametric Watson-Williams multi-sample test for equal means. Can be 
    used as a one-way ANOVA test for cicular data. Adapted for binned data."""        
    n1 = sum(w1)
    n2 = sum(w2)
    N = n1 + n2
    
    r1 = circ_r(alpha1, w1)    
    r2 = circ_r(alpha2, w2)
    
    r = circ_r(concatenate((alpha1, alpha2)), concatenate((w1, w2)))
    rw = sum((n1 * r1, n2 * r2)) / N
    
    if warnings:
        # check assumptions
        if N >= 11 and rw < .45:
            print('Warning: Test not applicable.'
                  'Average resultant vector length < 0.45.')
        elif N < 11 and N >= 7 and rw < .5:
            print('Test not applicable. Average number of samples per population '
                  '6 < x < 11 and average resultant vector length < 0.5.')
        elif N >= 5 and N < 7 and rw < .55:
            print('Test not applicable. Average number of samples per population '
                  '4 < x < 7 and average resultant vector length < 0.55.')
        elif N < 5:
            print('Test not applicable. '
                  'Average number of samples per population < 5.')
    
    # test statistic
    kk = circ_kappa(rw)
    beta = 1 + 3 / (8 * kk) # correction factor
    A = sum((n1 * r1, n2 * r2)) - r * N
    B = N - sum((n1 * r1, n2 * r2))
    
    F = beta * (N - 2) *  A / B
    pval = 1 - fdist.cdf(F, 1, N - 2)
    
    return F, pval
        
def circ_kappa(alpha, w=None):
    """ Computes an approximation to the ML estimate of the concentration 
    parameter kappa of the von Mises distribution."""    
    if isinstance(alpha, float):
        r = alpha
        N = 1        
    else:
        r = circ_r(alpha, w)
        N = len(alpha)
        
    if r < .53:
        kappa = (2 * r) + (r ** 3) + (5 * r ** (5/6))
    elif r >= .53 and r < .85:
        kappa = -.4 + 1.39 * r + .43 / (1 - r)
    else:
        kappa = 1 / (r ** 3 - 4 * r ** 2 + 3 * r)
        
    if N < 15 and N > 1:
        if kappa < 2:
            kappa = max(kappa - 2 / (N * kappa), 0)
        else:
            kappa = (N - 1) ** 3 * kappa / (N ** 3 + N)
            
    return kappa

def polar_plot(w, bin_edges):
    nbins = len(bin_edges) - 1
    width = 2 * pi / nbins
    f, ax = plt.subplots(1, 1, subplot_kw=dict(projection='polar'))
    ax.bar(bin_edges[:nbins], w, width=width, bottom=0.0)
    xL = ['0', '', '', '', r'+/- $\pi$', '', '', '']
    ax.set_xticklabels(xL)
    ax.set_yticks([])
    f.show()

def mean_amp(phase, amp, nbins=18):
    # Meanamp taking Hilbert transformed phase and amplitude signals (for comodulogram)
    # phase is the phase time series, type = array
    # amp is the amplitude time series, type = array
    # position is the array of phase bin left boundaries, type = array   
    # First, we break 2pi into phase bins
    # 0 degrees corresponds to the negative trough of the Phase freq
    # 180 degrees corresponds to the positive peak of the Phase freq       
    position = zeros(nbins)
    winsize = 2 * pi / nbins
    for i in range(nbins):
        position[i] = -pi + i * winsize
    position = roll(position, int(floor(nbins/4)-1), axis=-1)
    # Then, we compute mean amplitude in each phase bin:
    ma = zeros(nbins)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        for i, j in enumerate(position):
            ma[i] = nanmean(amp[where(
                logical_and(phase < (j + winsize), phase >= j)
            )])        
    return ma

def klentropy(MeanAmp, axis=-1):    
    # Quantify the amount of amp modulation by means of a normalized 
    # Kullback-Leibler entropy index
    nbin = MeanAmp.shape[-1]
    MI = (log(nbin) - (-sum((MeanAmp / sum(MeanAmp, axis=axis, keepdims=True)) * \
        log((MeanAmp / sum(MeanAmp, axis=axis, keepdims=True))), axis=axis, 
        keepdims=True))) / log(nbin)
    return MI.squeeze()

# =============================================================================
# ALL NIGHT AMPBIN WITH STD DEV
# def _allnight_ampbin(a, nodat, nbins, norm=True, sd=False):
#     out = zeros((*a.shape[:-2], nbins))
#     out_sd = zeros((*a.shape[:-2], nbins))
#     for i in range(a.shape[0]): # part
#         for j in range(a.shape[1]): # night
#             allcyc = [x for x in a[i, j, 0, :] if x is not nodat]
#             allcyc = concatenate(allcyc)
#             if norm:
#                 allcyc /= allcyc.sum(axis=1, keepdims=True)
#             out[i, j, :] = nanmean(allcyc, axis=0)
#             if std:
#                 out_sd[i, j, :] = nanstd(allcyc, axis=0)
#                 out = (out, out_sd)                
#     return out
#
# OLD _allnight_ampbin
# def _allnight_ampbin(a, nodat, nbins, norm=True):
#     out = zeros((*a.shape[:-2], nbins))
#     for i in range(a.shape[0]): # part
#         for j in range(a.shape[1]): # night
#             allcyc = [x for x in a[i, j, 0, :] if x is not nodat]
#             allcyc = concatenate(allcyc)
#             if norm:
#                 allcyc /= allcyc.sum(axis=1, keepdims=True)
#             out[i, j, :] = nanmean(allcyc, axis=0)                
#     return out
# =============================================================================

def _allnight_ampbin(a, nodat, nbins, norm=True):
    out = zeros((*a.shape[:-2], nbins))
    for i in range(a.shape[0]): # part
            allcyc = [x for x in a[i] if x is not nodat]
            allcyc = concatenate(allcyc)
            if norm:
                allcyc /= allcyc.sum(axis=1, keepdims=True)
            out[i, :] = nanmean(allcyc, axis=0)                
    return out

def ab_loader(ab_file, nbins, shift, norm=True):
    with open(ab_file, 'rb') as f:
        ab = load(f)
    try:
        ab = _allnight_ampbin(ab, 0, nbins, norm=norm)
    except(ValueError):
        ab = ab / ab.sum(-1, keepdims=True)
        ab = ab.squeeze()    
    ab = roll(ab, shift, axis=-1)
    
    return ab

def _idx_to_phase(idx, vecbin):
    phase = zeros(idx.shape, dtype='O')
    for i in range(idx.shape[0]):
        for j in range(idx.shape[1]):
            one_idx = asarray(idx[i][j], dtype='int')
            phase[i][j] = vecbin[one_idx]
            
    return phase

def logsd(m, sd):
    print('M: ' + str(around(exp(m), 3)))
    print('SD: ' + str(around(sqrt((exp(sd ** 2) - 1) * exp(2 * m + sd ** 2)), 3)))
    
def logse(m, sd, N=15):
    print('M: ' + str(around(exp(m), 3)))
    print('SE: ' + str(around(
            sqrt((exp(sd ** 2) - 1) * exp(2 * m + sd ** 2)) / sqrt(N), 3)))

def cohend(x, y):
    mean_diff = x.mean() - y.mean()
    pooled_sd = sqrt((x.std() ** 2 + y.std() ** 2) / 2)
    return mean_diff / pooled_sd

    
def paplot(ma):
    f, axarr = plt.subplots(nrows=2, sharex=True)
    width = 2 * pi / len(ma)
    pos = arange(0, 4 * pi, width)
    x = linspace(0, 4 * pi, 1000)
    pa = axarr[0]
    pa.bar(pos, hstack((ma, ma)), width=width, color='k', edgecolor='w')
    pa.set_xticks([0, pi/2, pi, 3 * pi / 2, 2 * pi, 5 * pi / 2, 3 * pi, 7 * pi / 2, 4 * pi])
    pa.set_xticklabels(['0'] + [r'$\pi$/2', r'$\pi$', r'3$\pi$/2', r'2$\pi$'] * 2)
    sine = axarr[1]
    sine.plot(sin(x))
    sine.set_xticks([])
    sine.set_yticks([])
    plt.show()
    


"""
SEA main module
"""

__author__ = 'Nathan Cross'
__credits__ = 'Concordia University (Canada), University of Sydney (Australia)'

from .version import __version__

from seapipe.core.dataset import pipeline




